"""SKALE test utilities"""

from contextlib import contextmanager
from timeit import default_timer as timer
from unittest.mock import MagicMock, Mock

from eth_typing import HexStr
from web3 import Web3

from skale import FairManager, SkaleAllocator, SkaleManager
from skale.utils.helper import get_allocator_address, get_skale_manager_address
from skale.wallets import Web3Wallet
from tests.constants import (
    ENDPOINT,
    ETH_PRIVATE_KEY,
    FAIR_CONTRACTS,
    TEST_ABI_FILEPATH,
    TEST_ALLOCATOR_ABI_FILEPATH,
)


def response_mock(status_code=0, json_data=None, cookies=None, headers=None, raw=None):
    result = MagicMock()
    result.status_code = status_code
    result.json = MagicMock(return_value=json_data)
    result.cookies = cookies
    result.headers = headers
    result.raw = raw
    return result


def request_mock(response_mock):
    return Mock(return_value=response_mock)


def init_skale(
    web3: Web3,
    eth_private_key: HexStr = ETH_PRIVATE_KEY,
    test_abi_filepath: str = TEST_ABI_FILEPATH,
) -> SkaleManager:
    wallet = Web3Wallet(eth_private_key, web3)
    return SkaleManager(
        ENDPOINT, get_skale_manager_address(test_abi_filepath), wallet, enable_stats=True
    )


def init_fair(web3: Web3, eth_private_key: HexStr = ETH_PRIVATE_KEY) -> FairManager:
    wallet = Web3Wallet(eth_private_key, web3)
    if not FAIR_CONTRACTS:
        raise ValueError('FAIR_CONTRACTS is not set')
    return FairManager(ENDPOINT, FAIR_CONTRACTS, wallet)


def init_skale_allocator(
    web3: Web3,
    eth_private_key: HexStr = ETH_PRIVATE_KEY,
    test_allocator_abi_filepath: str = TEST_ALLOCATOR_ABI_FILEPATH,
) -> SkaleAllocator:
    wallet = Web3Wallet(eth_private_key, web3)
    return SkaleAllocator(ENDPOINT, get_allocator_address(test_allocator_abi_filepath), wallet)


@contextmanager
def in_time(seconds):
    start_ts = timer()
    yield
    ts_diff = timer() - start_ts
    assert ts_diff < seconds, (ts_diff, seconds)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


from time import sleep
from typing import Any

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.types import Wei

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.wallets import BaseWallet

ERC20_ABI = [
    {
        'inputs': [],
        'name': 'name',
        'outputs': [{'name': '', 'type': 'string'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    {
        'inputs': [],
        'name': 'symbol',
        'outputs': [{'name': '', 'type': 'string'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    {
        'inputs': [],
        'name': 'decimals',
        'outputs': [{'name': '', 'type': 'uint8'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    {
        'inputs': [],
        'name': 'totalSupply',
        'outputs': [{'name': '', 'type': 'uint256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    {
        'inputs': [{'name': 'account', 'type': 'address'}],
        'name': 'balanceOf',
        'outputs': [{'name': '', 'type': 'uint256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    {
        'inputs': [{'name': 'to', 'type': 'address'}, {'name': 'amount', 'type': 'uint256'}],
        'name': 'transfer',
        'outputs': [{'name': '', 'type': 'bool'}],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    {
        'inputs': [{'name': 'owner', 'type': 'address'}, {'name': 'spender', 'type': 'address'}],
        'name': 'allowance',
        'outputs': [{'name': '', 'type': 'uint256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    {
        'inputs': [{'name': 'spender', 'type': 'address'}, {'name': 'amount', 'type': 'uint256'}],
        'name': 'approve',
        'outputs': [{'name': '', 'type': 'bool'}],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    {
        'inputs': [
            {'name': 'from', 'type': 'address'},
            {'name': 'to', 'type': 'address'},
            {'name': 'amount', 'type': 'uint256'},
        ],
        'name': 'transferFrom',
        'outputs': [{'name': '', 'type': 'bool'}],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
]


class Erc20Contract(BaseContract):
    def __init__(
        self,
        web3: Web3,
        address: ChecksumAddress,
        wallet: BaseWallet | None = None,
        erc20_abi: list[Any] | None = ERC20_ABI,
    ):
        abi_to_use = erc20_abi if erc20_abi is not None else ERC20_ABI

        super().__init__(web3, address, abi_to_use, wallet)

    def balance_of(self, account: ChecksumAddress) -> Wei:
        return Wei(self.contract.functions.balanceOf(account).call())

    @transaction_method
    def transfer(self, to: ChecksumAddress, amount: Wei) -> ContractFunction:
        return self.contract.functions.transfer(to, amount)

    @transaction_method
    def approve(self, spender: ChecksumAddress, amount: Wei) -> ContractFunction:
        return self.contract.functions.approve(spender, amount)

    def wait_for_balance_change(
        self,
        address: ChecksumAddress,
        initial_balance: Wei,
        timeout: int = 120,
        poll_interval: int = 5,
        raise_on_timeout: bool = False,
    ) -> bool:
        iterations = timeout // poll_interval + 1
        for _ in range(iterations):
            current_balance = self.balance_of(address)
            if current_balance != initial_balance:
                return True
            sleep(poll_interval)

        if raise_on_timeout:
            raise TimeoutError(f'Balance did not change for {address} within {timeout} seconds')
        return False



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import binascii
import json
import logging
import os
import time
from enum import Enum
from typing import Optional, Tuple, TypedDict

from eth_account.datastructures import SignedMessage, SignedTransaction
from eth_typing import ChecksumAddress, HexStr
from redis import Redis
from web3.types import TxParams, TxReceipt

import skale.config as config
from skale.transactions.exceptions import (
    TransactionError,
    TransactionNotMinedError,
    TransactionNotSentError,
    TransactionWaitError,
)
from skale.utils.web3_utils import DEFAULT_BLOCKS_TO_WAIT, MAX_WAITING_TIME
from skale.wallets import BaseWallet
from skale.wallets.web3_wallet import Web3Wallet

logger = logging.getLogger(__name__)


class RedisWalletError(Exception):
    pass


class RedisWalletDroppedError(RedisWalletError, TransactionNotMinedError):
    pass


class RedisWalletEmptyStatusError(RedisWalletError, TransactionError):
    pass


class RedisWalletNotSentError(RedisWalletError, TransactionNotSentError):
    pass


class RedisWalletWaitError(RedisWalletError, TransactionWaitError):
    pass


class TxRecordStatus(str, Enum):
    DROPPED = 'DROPPED'
    SUCCESS = 'SUCCESS'
    FAILED = 'FAILED'

    def __str__(self) -> str:
        return str.__str__(self)


TxRecord = TypedDict(
    'TxRecord',
    {
        'status': TxRecordStatus,
        'score': int,
        'multiplier': Optional[float],
        'tx_hash': HexStr,
        'tx_id': str | None,
        'method': Optional[str],
        'value': int,
        'chainId': int | None,
        'gas': int | None,
        'gasPrice': int | None,
        'from': ChecksumAddress | None,
        'to': ChecksumAddress,
        'maxFeePerGas': int | None,
        'maxPriorityFeePerGas': int | None,
        'hashes': list | None,
        'attempts': int | None,
        'source': str | None,
        'nonce': int | None,
        'data': dict | None,
        'meta': dict | None,
        'sent_ts': int | None,
    },
)


class RedisWalletAdapter(BaseWallet):
    ID_SIZE = 16

    def __init__(
        self,
        rs: Redis,
        pool: str,
        web3_wallet: Web3Wallet,
    ) -> None:
        self.rs = rs
        self.pool = pool
        self.wallet = web3_wallet

    def sign(self, tx: TxParams) -> SignedTransaction:
        return self.wallet.sign(tx)

    def sign_hash(self, unsigned_hash: str) -> SignedMessage:
        return self.wallet.sign_hash(unsigned_hash)

    @property
    def address(self) -> ChecksumAddress:
        return self.wallet.address

    @property
    def public_key(self) -> str:
        return self.wallet.public_key

    @classmethod
    def _make_raw_id(cls) -> bytes:
        prefix = b'tx-'
        unique = binascii.b2a_hex(os.urandom(cls.ID_SIZE // 2))
        return prefix + unique

    @classmethod
    def _make_score(cls, priority: int) -> int:
        ts = int(time.time())
        return priority * int(10 ** len(str(ts))) + ts

    @classmethod
    def _make_record(
        cls,
        tx: TxParams,
        score: int,
        multiplier: float = config.DEFAULT_GAS_MULTIPLIER,
        method: Optional[str] = None,
    ) -> Tuple[bytes, bytes]:
        tx_id = cls._make_raw_id()
        params = {
            'status': 'PROPOSED',
            'score': score,
            'multiplier': multiplier,
            'tx_hash': None,
            'method': method,
            **tx,
        }
        record = json.dumps(params).encode('utf-8')
        return tx_id, record

    @classmethod
    def _to_raw_id(cls, tx_id: str) -> bytes:
        return tx_id.encode('utf-8')

    @classmethod
    def _to_id(cls, raw_id: bytes) -> str:
        return raw_id.decode('utf-8')

    def sign_and_send(
        self,
        tx: TxParams,
        multiplier: Optional[float] = None,
        priority: Optional[int] = None,
        method: Optional[str] = None,
    ):
        priority = priority or config.DEFAULT_PRIORITY
        try:
            logger.info('Sending %s to redis pool, method: %s', tx, method)
            score = self._make_score(priority)
            raw_id, tx_record = self._make_record(
                tx, score, multiplier=multiplier or config.DEFAULT_GAS_MULTIPLIER, method=method
            )
            pipe = self.rs.pipeline()
            logger.info('Adding tx %s to the pool', raw_id)
            pipe.zadd(self.pool, {raw_id: score})
            logger.info('Saving tx %s record: %s', raw_id, tx_record)
            pipe.set(raw_id, tx_record, ex=config.TXRECORD_EXPIRATION)
            pipe.execute()
            return self._to_id(raw_id)
        except Exception as err:
            logger.exception(f'Sending {tx} with redis wallet errored')
            raise RedisWalletNotSentError(err)

    def get_status(self, tx_id: str) -> str:
        return self.get_record(tx_id)['status']

    def get_record(self, tx_id: str) -> TxRecord:
        rid = self._to_raw_id(tx_id)
        response = self.rs.get(rid)
        if isinstance(response, bytes):
            parsed_json = json.loads(response.decode('utf-8'))
            return TxRecord(
                {
                    'tx_id': parsed_json.get('tx_id'),
                    'status': parsed_json['status'],
                    'score': parsed_json['score'],
                    'multiplier': parsed_json.get('multiplier'),
                    'tx_hash': parsed_json['tx_hash'],
                    'method': parsed_json.get('method'),
                    'value': parsed_json['value'],
                    'chainId': parsed_json.get('chainId'),
                    'gas': parsed_json.get('gas'),
                    'gasPrice': parsed_json.get('gasPrice'),
                    'from': parsed_json.get('from'),
                    'to': parsed_json['to'],
                    'maxFeePerGas': parsed_json.get('maxFeePerGas'),
                    'maxPriorityFeePerGas': parsed_json.get('maxPriorityFeePerGas'),
                    'hashes': parsed_json.get('hashes'),
                    'attempts': parsed_json.get('attempts'),
                    'source': parsed_json.get('source'),
                    'nonce': parsed_json.get('nonce'),
                    'data': parsed_json.get('data'),
                    'meta': parsed_json.get('meta'),
                    'sent_ts': parsed_json.get('sent_ts'),
                }
            )
        raise ValueError('Unknown value was returned from get() call', response)

    def wait(
        self,
        tx_id,
        blocks_to_wait: int = DEFAULT_BLOCKS_TO_WAIT,
        timeout: int = MAX_WAITING_TIME,
    ) -> TxReceipt:
        start_ts = time.time()
        status, result = None, None
        while (
            status not in [TxRecordStatus.DROPPED, TxRecordStatus.SUCCESS, TxRecordStatus.FAILED]
            and time.time() - start_ts < timeout
        ):
            try:
                record = self.get_record(tx_id)
                if record is not None:
                    status = record.get('status')
                    if status in (TxRecordStatus.SUCCESS, TxRecordStatus.FAILED):
                        result = self.wallet._web3.eth.get_transaction_receipt(record['tx_hash'])
            except Exception as e:
                logger.exception('Waiting for tx %s errored', tx_id)
                raise RedisWalletWaitError(e)

        if result:
            return result

        if status is None:
            raise RedisWalletEmptyStatusError(f'Tx status is {status}')
        elif status == TxRecordStatus.DROPPED:
            raise RedisWalletDroppedError('Tx was dropped after max retries')
        else:
            raise RedisWalletWaitError(f'Tx finished with status {status}')



"""SKALE config test"""

import logging
from unittest import mock

import pytest
from web3.auto import w3

from skale import SkaleManager
from skale.types.node import NodeId
from skale.types.schain import SchainName
from skale.types.validator import ValidatorId
from skale.utils.account_tools import send_eth
from skale.utils.contracts_provision.fake_multisig_contract import deploy_fake_multisig_contract
from skale.utils.contracts_provision.main import (
    add_test2_schain_type,
    add_test_permissions,
    create_nodes,
    create_schain,
    link_nodes_to_validator,
    setup_validator,
)
from skale.utils.contracts_provision.utils import generate_random_name, generate_random_node_data
from skale.utils.helper import get_skale_manager_address
from skale.utils.web3_utils import init_web3
from skale.wallets import Web3Wallet
from skale.wallets.web3_wallet import generate_wallet
from tests.constants import ENDPOINT, TEST_ABI_FILEPATH
from tests.helper import init_fair, init_skale, init_skale_allocator

ETH_AMOUNT_PER_NODE = 1
logger = logging.getLogger(__name__)

_last_finished_test_name: str | None = None


@pytest.fixture(scope='session')
def web3():
    """Returns a SKALE Manager instance with provider from config"""
    w3 = init_web3(ENDPOINT)
    return w3


@pytest.fixture(scope='session')
def skale(web3, request):
    """Returns a cached SKALE Manager instance with provider from config"""
    if not hasattr(request.config, '_cached_skale'):
        skale_obj = init_skale(web3)
        add_test_permissions(skale_obj)
        add_test2_schain_type(skale_obj)
        if skale_obj.constants_holder.get_launch_timestamp() != 0:
            skale_obj.constants_holder.set_launch_timestamp(0)
        deploy_fake_multisig_contract(skale_obj.web3, skale_obj.wallet)
        request.config._cached_skale = skale_obj
    return request.config._cached_skale


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    report = outcome.get_result()
    if report.when == 'call' and report.outcome in {'passed', 'failed', 'skipped'}:
        item.config._last_finished_test_name = item.name


@pytest.fixture(scope='session')
def _skale_stats_logger(skale):
    def _log(test_name: str) -> None:
        logger.warning(
            test_name
            + ' '
            + str(skale.stats.http_posts)
            + ' '
            + str(skale.stats.by_method.most_common(10))
        )

    return _log


@pytest.fixture(autouse=True)
def log_prev_test_stats(request, _skale_stats_logger):
    global _last_finished_test_name
    prev_name = getattr(request.config, '_last_finished_test_name', None)
    if prev_name is not None:
        _skale_stats_logger(prev_name)
    _last_finished_test_name = request.node.name


@pytest.fixture(scope='session')
def fair(web3):
    return init_fair(web3)


@pytest.fixture(scope='session')
def validator(skale) -> ValidatorId:
    return setup_validator(skale)


@pytest.fixture(scope='session')
def number_of_nodes():
    return 2


@pytest.fixture
def node_wallets(skale: SkaleManager, number_of_nodes: int):
    wallets = []
    for _ in range(number_of_nodes):
        wallet = generate_wallet(skale.web3)
        send_eth(
            web3=skale.web3,
            wallet=skale.wallet,
            receiver_address=wallet.address,
            amount=ETH_AMOUNT_PER_NODE,
        )
        wallets.append(wallet)
    return wallets


@pytest.fixture
def node_skales(skale, node_wallets):
    return [
        SkaleManager(ENDPOINT, get_skale_manager_address(TEST_ABI_FILEPATH), wallet)
        for wallet in node_wallets
    ]


@pytest.fixture
def nodes(skale: SkaleManager, node_skales, validator: ValidatorId) -> list[NodeId]:
    link_nodes_to_validator(skale, validator, node_skales)
    ids = create_nodes(node_skales)
    return ids


@pytest.fixture
def node(skale: SkaleManager, validator: ValidatorId) -> tuple[NodeId, Web3Wallet]:
    wallet = generate_wallet(skale.web3)
    send_eth(
        web3=skale.web3,
        wallet=skale.wallet,
        receiver_address=wallet.address,
        amount=ETH_AMOUNT_PER_NODE,
    )
    node_skales = (SkaleManager(ENDPOINT, get_skale_manager_address(TEST_ABI_FILEPATH), wallet),)
    link_nodes_to_validator(skale, validator, node_skales)
    ids = create_nodes(node_skales, names=[generate_random_name()])
    node_id = ids[0]
    return node_id, wallet


@pytest.fixture
def fair_active_nodes(fair, node_wallets):
    main_wallet = fair.wallet

    for wallet in node_wallets:
        fair.wallet = wallet
        ip, _, port, _ = generate_random_node_data()
        self_stake_requirement = fair.staking.self_stake_requirement()
        fair.nodes.register_active(ip=ip, port=port, value=self_stake_requirement)

    fair.wallet = main_wallet
    try:
        yield node_wallets
    finally:
        """TODO: Remove the node from the fair instance."""


@pytest.fixture
def fair_passive_nodes(fair, node_wallets):
    main_wallet = fair.wallet

    for wallet in node_wallets:
        fair.wallet = wallet
        ip, _, port, _ = generate_random_node_data()
        fair.nodes.register_passive(ip=ip, port=port)

    fair.wallet = main_wallet
    try:
        yield node_wallets
    finally:
        """TODO: Remove the node from the fair instance."""


@pytest.fixture
def schain(skale: SkaleManager, nodes: list[NodeId]) -> SchainName:
    return create_schain(
        skale,
        schain_type=1,  # test2 should have 1 index
        random_name=True,
    )


@pytest.fixture(autouse=True)
def isolation(web3):
    response = web3.provider.make_request('evm_snapshot', [])
    if 'result' not in response:
        logger.warning('EVM snapshot not supported by the provider')
        yield
        return
    snapshot_id = response['result']
    try:
        yield
    finally:
        web3.provider.make_request('evm_revert', [snapshot_id])


@pytest.fixture(scope='session')
def skale_allocator(web3):
    """Returns a SKALE Allocator instance with provider from config"""
    return init_skale_allocator(web3)


@pytest.fixture
def empty_account():
    return w3.eth.account.create()


@pytest.fixture
def failed_skale(skale):
    tmp_wait, tmp_sign_and_send = skale.wallet.wait, skale.wallet.sign_and_send
    skale.wallet.sign_and_send = mock.Mock(return_value='0x000000000')
    skale.wallet.wait = mock.Mock(return_value={'status': 0, 'error': 'Test error'})
    try:
        yield skale
    finally:
        skale.wallet.sign_and_send = tmp_sign_and_send
        skale.wallet.wait = tmp_wait



"""Tests for contracts/delegation/validator_service.py"""

import random

import pytest

from skale.contracts.manager.delegation.validator_service import FIELDS
from skale.transactions.exceptions import TransactionFailedError
from skale.transactions.result import DryRunRevertError
from skale.utils.account_tools import send_eth
from skale.utils.contracts_provision.main import _skip_evm_time, enable_validator
from skale.wallets.web3_wallet import generate_wallet
from tests.constants import (
    D_DELEGATION_PERIOD,
    D_VALIDATOR_DESC,
    D_VALIDATOR_FEE,
    D_VALIDATOR_ID,
    D_VALIDATOR_MIN_DEL,
    D_VALIDATOR_NAME,
    MONTH_IN_SECONDS,
    NOT_EXISTING_ID,
)


def link_node_address(skale, wallet, validator_id=D_VALIDATOR_ID):
    main_wallet = skale.wallet
    skale.wallet = wallet
    try:
        signature = skale.validator_service.get_link_node_signature(validator_id=validator_id)
    finally:
        skale.wallet = main_wallet

    skale.validator_service.link_node_address(
        node_address=wallet.address, signature=signature, wait_for=True
    )


def _generate_new_validator(skale, wallet=None, enable=True):
    eth_amount = 10
    main_wallet = skale.wallet
    wallet = wallet or generate_wallet(skale.web3)
    send_eth(skale.web3, main_wallet, wallet.address, eth_amount)
    validator_id = -1
    try:
        skale.wallet = wallet
        skale.validator_service.register_validator(
            name=D_VALIDATOR_NAME,
            description=D_VALIDATOR_DESC,
            fee_rate=D_VALIDATOR_FEE,
            min_delegation_amount=D_VALIDATOR_MIN_DEL,
        )
        validator_id = skale.validator_service.validator_id_by_address(wallet.address)
    finally:
        skale.wallet = main_wallet

    assert validator_id > 0, validator_id
    if enable:
        enable_validator(skale, validator_id)
    return validator_id


def test_get_raw_not_exist(skale):
    empty_struct = skale.validator_service._ValidatorService__get_raw(NOT_EXISTING_ID)
    assert empty_struct[0] == ''
    assert empty_struct[1] == '0x0000000000000000000000000000000000000000'


def test_get(skale, validator):
    validator_id = validator
    validator_data = skale.validator_service.get(validator_id)
    assert list(validator_data.keys()) == FIELDS
    assert [k for k, v in validator_data.items() if v is None] == []


def test_get_with_id(skale):
    validator = skale.validator_service.get_with_id(D_VALIDATOR_ID)
    assert validator['id'] == D_VALIDATOR_ID


def test_number_of_validators(skale):
    n_of_validators_before = skale.validator_service.number_of_validators()
    _generate_new_validator(skale)
    n_of_validators_after = skale.validator_service.number_of_validators()
    assert n_of_validators_after == n_of_validators_before + 1


def test_ls(skale, validator):
    n_of_validators = skale.validator_service.number_of_validators()
    validators = sorted(skale.validator_service.ls(), key=lambda x: x['validator_address'])
    assert n_of_validators == len(validators)
    session_validator_data = skale.validator_service.get_with_id(validator)
    assert session_validator_data in validators
    trusted_validators = sorted(
        skale.validator_service.ls(trusted_only=True), key=lambda x: x['validator_address']
    )
    assert all(v['trusted'] for v in trusted_validators)
    assert list(filter(lambda v: v['trusted'], validators)) == trusted_validators


def test_get_linked_addresses_by_validator_address(skale, validator):
    addresses = skale.validator_service.get_linked_addresses_by_validator_address(
        address=skale.wallet.address
    )

    wallet = generate_wallet(skale.web3)
    link_node_address(skale, wallet, validator_id=validator)

    assert wallet.address not in addresses
    addresses_by_address = skale.validator_service.get_linked_addresses_by_validator_address(
        address=skale.wallet.address
    )
    assert wallet.address in addresses_by_address
    addresses_by_id = skale.validator_service.get_linked_addresses_by_validator_id(
        validator_id=validator
    )
    assert addresses_by_address == addresses_by_id


def test_is_main_address(skale, validator):
    is_main_address = skale.validator_service.is_main_address(skale.wallet.address)
    assert is_main_address

    wallet = generate_wallet(skale.web3)
    link_node_address(skale, wallet, validator_id=validator)

    is_main_address = skale.validator_service.is_main_address(wallet.address)
    assert not is_main_address


def test_validator_address_exists(skale):
    address_exists = skale.validator_service.validator_address_exists(skale.wallet.address)
    assert address_exists

    wallet = generate_wallet(skale.web3)
    address_exists = skale.validator_service.validator_address_exists(wallet.address)
    assert not address_exists


def test_validator_exists(skale, validator):
    validator_exists = skale.validator_service.validator_exists(validator_id=validator)
    assert validator_exists
    validator_exists = skale.validator_service.validator_exists(NOT_EXISTING_ID)
    assert not validator_exists


def test_validator_node_indices(skale, validator):  # todo: improve test
    node_indices = skale.nodes.validator_node_indices(validator_id=validator)
    all_active_node_ids = skale.nodes.active_node_ids()
    assert set(all_active_node_ids).issubset(node_indices)


def test_enable_validator(skale):
    validator_id = _generate_new_validator(skale, enable=False)
    is_validator_trusted = skale.validator_service._is_authorized_validator(
        validator_id=validator_id
    )
    assert not is_validator_trusted

    skale.validator_service._enable_validator(validator_id=validator_id)

    is_validator_trusted = skale.validator_service._is_authorized_validator(
        validator_id=validator_id
    )
    assert is_validator_trusted


def test_disable_validator(skale):
    validator_id = _generate_new_validator(skale, enable=True)

    is_validator_trusted = skale.validator_service._is_authorized_validator(
        validator_id=validator_id
    )
    assert is_validator_trusted

    skale.validator_service._disable_validator(validator_id=validator_id)
    is_validator_trusted = skale.validator_service._is_authorized_validator(validator_id)
    assert not is_validator_trusted


def test_is_accepting_new_requests(skale, validator):
    is_accepting_new_requests = skale.validator_service.is_accepting_new_requests(
        validator_id=validator
    )
    assert is_accepting_new_requests


def test_register_existing_validator(skale):
    with pytest.raises(DryRunRevertError):
        skale.validator_service.register_validator(
            name=D_VALIDATOR_NAME,
            description=D_VALIDATOR_DESC,
            fee_rate=D_VALIDATOR_FEE,
            min_delegation_amount=D_VALIDATOR_MIN_DEL,
            wait_for=True,
        )


def test_register_new_validator(skale):
    n_of_validators_before = skale.validator_service.number_of_validators()
    _generate_new_validator(skale)
    n_of_validators_after = skale.validator_service.number_of_validators()
    assert n_of_validators_after == n_of_validators_before + 1


def test_unlink_node_address(skale):
    wallet = generate_wallet(skale.web3)
    link_node_address(skale, wallet)

    addresses = skale.validator_service.get_linked_addresses_by_validator_address(
        skale.wallet.address
    )
    assert wallet.address in addresses

    skale.validator_service.unlink_node_address(
        node_address=wallet.address,
    )
    addresses = skale.validator_service.get_linked_addresses_by_validator_address(
        skale.wallet.address
    )
    assert wallet.address not in addresses


def test_get_and_update_bond_amount(skale, validator):
    _skip_evm_time(skale.web3, MONTH_IN_SECONDS * 3)
    initial_bond = skale.validator_service.get_and_update_bond_amount(validator_id=validator)
    additional_bond = skale.constants_holder.msr() * 2

    # Delegate to myself
    skale.delegation_controller.delegate(
        validator_id=validator,
        amount=additional_bond,
        delegation_period=D_DELEGATION_PERIOD,
        info='Test get_and_update_bond_amount',
        wait_for=True,
    )

    # Accept delegation
    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator
    )
    skale.delegation_controller.accept_pending_delegation(
        delegation_id=delegations[-1]['id'], wait_for=True
    )

    initial_bts = skale.web3.eth.get_block('latest').timestamp
    # Skip time
    _skip_evm_time(skale.web3, MONTH_IN_SECONDS)
    bts = skale.web3.eth.get_block('latest').timestamp

    bond = skale.validator_service.get_and_update_bond_amount(validator_id=validator)
    locked = skale.token_state.get_and_update_locked_amount(skale.wallet.address)
    assert bond == initial_bond + additional_bond, (
        bond,
        initial_bond + additional_bond,
        locked,
        initial_bts,
        bts,
    )


def test_set_validator_mda(skale, validator):
    validator_id = validator
    minimum_delegation_amount = random.randint(1000, 10000)

    validator = skale.validator_service.get(validator_id)
    old_mda = validator['minimum_delegation_amount']

    skale.validator_service.set_validator_mda(
        minimum_delegation_amount=minimum_delegation_amount, wait_for=True
    )
    validator = skale.validator_service.get(validator_id)
    new_mda = validator['minimum_delegation_amount']

    assert minimum_delegation_amount != old_mda
    assert minimum_delegation_amount == new_mda


def test_request_confirm_new_address(skale):
    main_wallet = skale.wallet
    try:
        old_wallet = generate_wallet(skale.web3)
        validator_id = _generate_new_validator(skale, wallet=old_wallet)
        skale.wallet = old_wallet
        new_wallet = generate_wallet(skale.web3)
        eth_amount = 10
        skale.wallet = main_wallet
        send_eth(skale.web3, skale.wallet, new_wallet.address, eth_amount)
        validator = skale.validator_service.get(validator_id)
        assert validator['requested_address'] == '0x0000000000000000000000000000000000000000'

        with pytest.raises(DryRunRevertError):
            skale.validator_service.request_for_new_address(
                new_validator_address=main_wallet.address, wait_for=True
            )
        skale.wallet = old_wallet
        skale.validator_service.request_for_new_address(
            new_validator_address=new_wallet.address, wait_for=True
        )
        validator = skale.validator_service.get(validator_id)
        assert validator['requested_address'] == new_wallet.address
        assert validator['validator_address'] == old_wallet.address

        skale.wallet = new_wallet
        skale.validator_service.confirm_new_address(validator_id=validator_id, wait_for=True)
        validator = skale.validator_service.get(validator_id)
        assert validator['requested_address'] == '0x0000000000000000000000000000000000000000'
        assert validator['validator_address'] == new_wallet.address
    finally:
        skale.wallet = main_wallet


def test_set_validator_name(skale):
    main_wallet = skale.wallet
    wallet = generate_wallet(skale.web3)
    validator_id = _generate_new_validator(skale, wallet=wallet)
    try:
        skale.wallet = wallet
        new_test_name = 'test_123'
        validator = skale.validator_service.get(validator_id)
        assert validator['name'] == D_VALIDATOR_NAME
        assert validator['name'] != new_test_name

        skale.validator_service.set_validator_name(new_name=new_test_name, wait_for=True)

        validator = skale.validator_service.get(validator_id)
        assert validator['name'] != D_VALIDATOR_NAME
        assert validator['name'] == new_test_name
    finally:
        skale.wallet = main_wallet


def test_set_validator_description(skale):
    main_wallet = skale.wallet
    wallet = generate_wallet(skale.web3)
    validator_id = _generate_new_validator(skale, wallet=wallet)
    try:
        skale.wallet = wallet
        new_test_description = 'test_description'
        validator = skale.validator_service.get(validator_id)
        assert validator['description'] == D_VALIDATOR_DESC
        assert validator['description'] != new_test_description

        skale.validator_service.set_validator_description(
            new_description=new_test_description, wait_for=True
        )

        validator = skale.validator_service.get(validator_id)
        assert validator['description'] != D_VALIDATOR_DESC
        assert validator['description'] == new_test_description
    finally:
        skale.wallet = main_wallet


def test_revert_register(skale):
    with pytest.raises(TransactionFailedError):
        skale.validator_service.register_validator(
            name=D_VALIDATOR_NAME,
            description=D_VALIDATOR_DESC,
            fee_rate=D_VALIDATOR_FEE,
            min_delegation_amount=D_VALIDATOR_MIN_DEL,
            wait_for=True,
            skip_dry_run=True,
        )


def test_get_use_whitelist(skale):
    assert skale.validator_service.get_use_whitelist()


@pytest.mark.skip('Breaks whitelist related tests')
def test_disable_whitelist(skale):
    assert skale.validator_service.get_use_whitelist()
    skale.validator_service.disable_whitelist(wait_for=True)
    assert not skale.validator_service.get_use_whitelist()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from functools import wraps
from typing import Any, Callable, Tuple, TypedDict

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction
from web3.types import Wei

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.types.validator import ValidatorId


class EarnedData(TypedDict):
    earned: Wei
    end_month: int


def formatter(method: Callable[..., Tuple[Wei, int]]) -> Callable[..., EarnedData]:
    @wraps(method)
    def wrapper(self: SkaleManagerContract, *args: Any, **kwargs: Any) -> EarnedData:
        res = method(self, *args, **kwargs)
        return EarnedData(
            {
                'earned': res[0],
                'end_month': res[1],
            }
        )

    return wrapper


class Distributor(SkaleManagerContract):
    """Wrapper for Distributor.sol functions"""

    @formatter
    def get_earned_bounty_amount(
        self, validator_id: ValidatorId, address: ChecksumAddress
    ) -> Tuple[Wei, int]:
        """Get earned bounty amount for the validator

        :param validator_id: ID of the validator
        :type validator_id: int
        :returns: Earned bounty amount and end month
        :rtype: dict
        """
        return tuple(
            self.contract.functions.getAndUpdateEarnedBountyAmount(validator_id).call(
                {'from': address}
            )
        )

    @formatter
    def get_earned_fee_amount(self, address: ChecksumAddress) -> Tuple[Wei, int]:
        """Get earned fee amount for the address

        :param address: Address of the validator
        :type address: ChecksumAddress
        :returns: Earned bounty amount and end month
        :rtype: dict
        """
        return tuple(self.contract.functions.getEarnedFeeAmount().call({'from': address}))

    @transaction_method
    def withdraw_bounty(self, validator_id: ValidatorId, to: ChecksumAddress) -> 'ContractFunction':
        """Withdraw earned bounty to specified address

        :param validator_id: ID of the validator
        :type validator_id: int
        :param to: Address to transfer bounty
        :type to: str
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.withdrawBounty(validator_id, to)

    @transaction_method
    def withdraw_fee(self, to: ChecksumAddress) -> 'ContractFunction':
        """Withdraw earned fee to specified address

        :param to: Address to transfer bounty
        :type to: str
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.withdrawFee(to)



from eth_account.datastructures import SignedMessage, SignedTransaction
from hexbytes import HexBytes
from sgx.sgx import Account

from skale.utils.web3_utils import (
    private_key_to_address,
    private_key_to_public,
    to_checksum_address,
)
from tests.constants import ETH_PRIVATE_KEY

ADDRESS = to_checksum_address(private_key_to_address(ETH_PRIVATE_KEY))
PUBLIC_KEY = private_key_to_public(ETH_PRIVATE_KEY)


class SgxClient:
    def __init__(self, endpoint, path_to_cert=None):
        pass

    def generate_key(self):
        return Account(name='NEK:aaabbb', address=ADDRESS, public_key=PUBLIC_KEY)

    def get_account(self, key_name):
        return Account(name='NEK:aaabbb', address=ADDRESS, public_key=PUBLIC_KEY)

    def sign(self, transaction_dict, key_name):
        return SignedTransaction(
            raw_transaction=HexBytes('0x000000000000'),
            hash=HexBytes('0x000000000000'),
            r=100000000000,
            s=100000000000,
            v=37,
        )

    def sign_hash(self, message, key_name, chain_id):
        return SignedMessage(
            message_hash=HexBytes('0x31323331'),
            r=123,
            s=123,
            v=27,
            signature=HexBytes('0x6161616161613131313131'),
        )


class BadSgxClient:
    def __init__(self, endpoint, path_to_cert=None):
        pass

    def generate_key(self):
        return Account(name='NEK:aaabbb', address=ADDRESS, public_key=PUBLIC_KEY)

    def get_account(self, key_name):
        return Account(name='NEK:aaabbb', address=ADDRESS, public_key=PUBLIC_KEY)

    def sign(self, transaction_dict, key_name):
        raise ValueError('Test Error')

    def sign_hash(self, message, key_name, chain_id):
        raise ValueError('Test error')



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2024-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Any, NewType

from eth_typing import ChecksumAddress
from skale_core.types import SchainName
from web3.types import Wei

from skale.dataclasses.schain_options import SchainOptions

SchainHash = NewType('SchainHash', bytes)
SchainOption = tuple[str, bytes]


@dataclass
class Schain:
    name: SchainName
    mainnet_owner: ChecksumAddress
    index_in_owner_list: int
    part_of_node: int
    lifetime: int
    start_date: int
    start_block: int
    deposit: Wei
    index: int
    generation: int
    originator: ChecksumAddress

    def to_dict(self) -> dict[str, Any]:
        return {
            'name': self.name,
            'mainnet_owner': self.mainnet_owner,
            'index_in_owner_list': self.index_in_owner_list,
            'part_of_node': self.part_of_node,
            'lifetime': self.lifetime,
            'start_date': self.start_date,
            'start_block': self.start_block,
            'deposit': int(self.deposit),
            'index': self.index,
            'generation': self.generation,
            'originator': self.originator,
        }


@dataclass
class SchainStructure(Schain):
    schain_hash: SchainHash
    options: SchainOptions
    active: bool

    def to_dict(self) -> dict[str, Any]:
        return {
            **super().to_dict(),
            'schain_hash': self.schain_hash.hex(),
            'options': {
                'multitransaction_mode': self.options.multitransaction_mode,
                'threshold_encryption': self.options.threshold_encryption,
                'allocation_type': self.options.allocation_type.value,
                'external_gas_difficulty': self.options.external_gas_difficulty,
                'min_gas_price': self.options.min_gas_price,
                'max_gas_price': self.options.max_gas_price,
            },
            'active': self.active,
        }



from dataclasses import astuple, fields

from skale.skale_manager import SkaleManager
from skale.types.schain import SchainName
from skale.utils.helper import schain_name_to_hash
from tests.constants import (
    DEFAULT_SCHAIN_HASH,
    EMPTY_SCHAIN_ARR,
    MIN_NODES_IN_SCHAIN,
    SCHAIN_FIELDS,
)


def test_get_raw(skale):
    schain = skale.schains_internal.get_raw(DEFAULT_SCHAIN_HASH)
    assert len(SCHAIN_FIELDS) == len(fields(schain)) + 2  # +2 for chainId + options


def test_get_raw_not_exist(skale):
    not_exist_schain_id = schain_name_to_hash('unused_hash')
    schain_arr = skale.schains_internal.get_raw(not_exist_schain_id)
    assert list(astuple(schain_arr)) == EMPTY_SCHAIN_ARR


def test_number_of_schains(skale: SkaleManager, schain):
    assert skale.schains_internal.number_of_schains() == 1


def test_schain_list_size(skale: SkaleManager, schain, empty_account):
    list_size = skale.schains_internal.schain_list_size(skale.wallet.address)
    empty_list_size = skale.schains_internal.schain_list_size(empty_account.address)

    assert list_size != 0
    assert empty_list_size == 0


def test_schain_hash_by_index_for_owner(skale: SkaleManager, schain):
    schain_id = skale.schains_internal.schain_hash_by_index_for_owner(skale.wallet.address, 0)
    schain = skale.schains.get(schain_id)
    assert schain.mainnet_owner == skale.wallet.address


def test_node_ids_for_schain(skale: SkaleManager, schain):
    schain_node_ids = skale.schains_internal.node_ids_for_schain(schain)

    assert isinstance(schain_node_ids, list)
    assert len(schain_node_ids) >= MIN_NODES_IN_SCHAIN


def test_schain_hashes_for_node(skale: SkaleManager, nodes, schain):
    node_id = nodes[0]
    schain_hashes_for_node = skale.schains_internal.schain_hashes_for_node(node_id)

    assert isinstance(schain_hashes_for_node, list)
    assert len(schain_hashes_for_node) > 0


def test_is_schain_exist(skale: SkaleManager, schain):
    assert skale.schains_internal.is_schain_exist(schain)
    nonexistent_chain_name = SchainName('sfdgodifjgodsaf')
    assert not skale.schains_internal.is_schain_exist(nonexistent_chain_name)


def test_get_active_schain_hashes(skale: SkaleManager, nodes, schain):
    node_id = nodes[0]
    active_schains = skale.schains_internal.active_schain_hashes_for_node(node_id)

    assert isinstance(active_schains, list)
    assert len(active_schains) > 0


def test_connected_node_ids(skale: SkaleManager, nodes, schain):
    node_id = nodes[0]
    node_ids = skale.schains_internal.connected_node_ids(node_id)
    assert isinstance(node_ids, list)
    assert len(node_ids) >= 1


def test_get_current_generation(skale):
    current_generation = skale.schains_internal.current_generation()
    assert isinstance(current_generation, int)


def test_generation_manager_role(skale):
    test_address = skale.web3.eth.account.create().address
    role = skale.schains_internal.generation_manager_role()
    assert not skale.schains_internal.has_role(role, test_address)
    skale.schains_internal.grant_role(role, test_address)
    assert skale.schains_internal.has_role(role, test_address)


def test_new_generation(skale):
    current_generation = skale.schains_internal.current_generation()
    skale.schains_internal.new_generation()
    new_generation = skale.schains_internal.current_generation()
    assert current_generation + 1 == new_generation



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import json
from collections import Counter
from dataclasses import dataclass, field
from threading import Lock
from typing import Any

import requests


@dataclass(slots=True)
class RpcHttpStats:
    http_posts: int = 0
    rpc_objects: int = 0
    by_method: Counter[str] = field(default_factory=Counter)
    transport_errors: int = 0


class CountingSession(requests.Session):
    def __init__(self, stats: RpcHttpStats):
        super().__init__()
        self.stats = stats
        self._lock = Lock()

    def request(self, *args: Any, **kwargs: Any) -> requests.Response:
        method = str(args[0]) if args else str(kwargs.get('method', ''))
        if method.upper() == 'POST':
            n, methods = _count_jsonrpc(kwargs)
            with self._lock:
                self.stats.http_posts += 1
                self.stats.rpc_objects += n
                self.stats.by_method.update(methods)
            try:
                return super().request(*args, **kwargs)
            except Exception:
                with self._lock:
                    self.stats.transport_errors += 1
                raise

        return super().request(*args, **kwargs)


def _count_jsonrpc(kwargs: dict[str, Any]) -> tuple[int, list[str]]:
    payload = kwargs.get('json')
    if payload is None:
        data = kwargs.get('data')
        if isinstance(data, (bytes, bytearray)):
            data = data.decode('utf-8', errors='replace')
        if isinstance(data, str):
            try:
                payload = json.loads(data)
            except Exception:
                return 0, []

    if isinstance(payload, dict):
        method = payload.get('method')
        return 1, [method] if isinstance(method, str) else []

    if isinstance(payload, list):
        methods = [obj.get('method') for obj in payload if isinstance(obj, dict)]
        return len(payload), [method for method in methods if isinstance(method, str)]

    return 0, []



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import List

from eth_typing import ChecksumAddress

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_contract import SkaleContract
from skale.types.exit_request import ExitRequest, exit_request_from_tuple
from skale.types.node import NodeId


class Staking(SkaleContract):
    def get_node_share(self, node: NodeId) -> int:
        return self.contract.functions.getNodeShare(node).call()

    def get_node_total_stake(self, node: NodeId) -> int:
        return self.contract.functions.getNodeTotalStake(node).call()

    def get_node_fee_rate(self, node: NodeId) -> int:
        return self.contract.functions.getNodeFeeRate(node).call()

    def is_node_enabled(self, node: NodeId) -> bool:
        return self.contract.functions.isNodeEnabled(node).call()

    def get_delegators_to_node(self, node: NodeId) -> List[ChecksumAddress]:
        return self.contract.functions.getDelegatorsToNode(node).call()

    def get_delegators_to_node_count(self, node: NodeId) -> int:
        return self.contract.functions.getDelegatorsToNodeCount(node).call()

    def get_staked_amount(self) -> int:
        return self.contract.functions.getStakedAmount().call({'from': self.skale.wallet.address})

    def get_staked_to_node_amount(self, node: NodeId) -> int:
        return self.contract.functions.getStakedToNodeAmount(node).call(
            {'from': self.skale.wallet.address}
        )

    def get_staked_nodes(self) -> List[NodeId]:
        return self.contract.functions.getStakedNodes().call({'from': self.skale.wallet.address})

    def get_earned_fee_amount(self, node: NodeId) -> int:
        return self.contract.functions.getEarnedFeeAmount(node).call()

    def get_staked_amount_for(self, holder: ChecksumAddress) -> int:
        return self.contract.functions.getStakedAmountFor(holder).call()

    def get_staked_nodes_for(self, holder: ChecksumAddress) -> List[NodeId]:
        return self.contract.functions.getStakedNodesFor(holder).call()

    def get_staked_to_node_amount_for(self, node: NodeId, holder: ChecksumAddress) -> int:
        return self.contract.functions.getStakedToNodeAmountFor(node, holder).call()

    def get_exit_requests_count_for(self, user: ChecksumAddress) -> int:
        return self.contract.functions.getExitRequestsCountFor(user).call()

    def get_total_in_exit_queue(self, address: ChecksumAddress) -> int:
        return self.contract.functions.getTotalInExitQueueFor(address).call()

    def get_my_exit_requests_count(self) -> int:
        return self.contract.functions.getMyExitRequestsCount().call(
            {'from': self.skale.wallet.address}
        )

    def is_within_stake_limit(self, node: NodeId) -> bool:
        return self.contract.functions.isWithinStakeLimit(node).call()

    def get_exit_request(self, request_id: int) -> ExitRequest:
        data = self.contract.functions.getExitRequest(request_id).call()
        return exit_request_from_tuple(data)

    def get_unlocked_exit_request_for(self, user: ChecksumAddress, from_index: int) -> ExitRequest:
        data = self.contract.functions.getUnlockedExitRequestFor(user, from_index).call()
        return exit_request_from_tuple(data)

    def get_exit_request_at(self, user: ChecksumAddress, index: int) -> ExitRequest:
        data = self.contract.functions.getExitRequestAt(user, index).call()
        return exit_request_from_tuple(data)

    def get_exit_requests_for(self, address: ChecksumAddress) -> List[ExitRequest]:
        count = self.get_exit_requests_count_for(address)
        return [self.get_exit_request_at(address, i) for i in range(count)]

    def is_request_unlocked(self, request_id: int) -> bool:
        return self.contract.functions.isRequestUnlocked(request_id).call()

    def get_retrieving_delay(self) -> int:
        return self.contract.functions.getRetrievingDelay().call()

    def self_stake_requirement(self) -> int:
        return self.contract.functions.selfStakeRequirement().call()

    @transaction_method
    def stake(self, node: NodeId):
        return self.contract.functions.stake(node)

    @transaction_method
    def request_retrieve(self, node: NodeId, value: int):
        return self.contract.functions.requestRetrieve(node, value)

    @transaction_method
    def request_retrieve_all(self, node: NodeId):
        return self.contract.functions.requestRetrieveAll(node)

    @transaction_method
    def set_fee_rate(self, fee_rate: int):
        return self.contract.functions.setFeeRate(fee_rate)

    @transaction_method
    def request_fees(self, node: NodeId, amount: int):
        return self.contract.functions.requestFees(node, amount)

    @transaction_method
    def request_all_fees(self, node: NodeId):
        return self.contract.functions.requestAllFees(node)

    @transaction_method
    def request_send_fees(self, to: ChecksumAddress, amount: int):
        return self.contract.functions.requestSendFees(to, amount)

    @transaction_method
    def request_send_all_fees(self, to: ChecksumAddress):
        return self.contract.functions.requestSendAllFees(to)

    @transaction_method
    def claim_request(self, request_id: int):
        return self.contract.functions.claimRequest(request_id)

    @transaction_method
    def disable(self, node: NodeId):
        return self.contract.functions.disable(node)

    @transaction_method
    def enable(self, node: NodeId):
        return self.contract.functions.enable(node)

    @transaction_method
    def add_allowed_receiver(self, receiver: ChecksumAddress):
        return self.contract.functions.addAllowedReceiver(receiver)

    @transaction_method
    def remove_allowed_receiver(self, receiver: ChecksumAddress):
        return self.contract.functions.removeAllowedReceiver(receiver)

    @transaction_method
    def set_stake_limit(self, limit: int):
        return self.contract.functions.setStakeLimit(limit)

    @transaction_method
    def set_reward_wallet_reference(self, reward_wallet_reference: ChecksumAddress):
        return self.contract.functions.setRewardWalletReference(reward_wallet_reference)

    @transaction_method
    def set_self_stake_requirement(self, amount: int):
        return self.contract.functions.setSelfStakeRequirement(amount)

    @transaction_method
    def set_retrieving_delay(self, delay: int):
        return self.contract.functions.setRetrievingDelay(delay)

    def get_reward_wallet(self, node: NodeId) -> ChecksumAddress:
        return self.contract.functions.getRewardWallet(node).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2026-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale_core.settings import (
    SETTINGS_MAP,
    ActiveSettings,
    BaseNodeSettings,
    FairBaseSettings,
    FairSettings,
    InternalSettings,
    SkalePassiveSettings,
    SkaleSettings,
    TomlBaseSettings,
    _resolve_type,
    get_internal_settings,
    get_settings,
    write_internal_settings_file,
    write_node_settings_file,
)

__all__ = [
    'ActiveSettings',
    'BaseNodeSettings',
    'FairBaseSettings',
    'FairSettings',
    'InternalSettings',
    'SETTINGS_MAP',
    'SkalePassiveSettings',
    'SkaleSettings',
    'TomlBaseSettings',
    '_resolve_type',
    'get_internal_settings',
    'get_settings',
    'write_internal_settings_file',
    'write_node_settings_file',
]



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2026-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import secrets
from dataclasses import dataclass
from time import perf_counter

from eth_account import Account
from hexbytes import HexBytes
from web3 import HTTPProvider, Web3

MAX_NUMBER = (1 << 256) - 1
DEFAULT_DIFFICULTY = 1
DEFAULT_TARGET_GAS = 1_000_000
DEFAULT_MINE_TIMEOUT_SECONDS = 60
DEFAULT_RECEIPT_TIMEOUT_SECONDS = 240

FAUCET_BY_CHAIN_ID: dict[int, tuple[str, str]] = {
    1_444_673_419: ('0x366727B410fE55774C8b0B5b5A6E2d74199a088A', '0x0c11dedd'),
}


@dataclass(frozen=True)
class MiningResult:
    success: bool
    tx_hash: HexBytes | None = None
    receipt_status: int | None = None
    gas_price: int | None = None
    iterations: int | None = None
    elapsed_ms: float | None = None
    start_balance: int | None = None
    end_balance: int | None = None
    balance_delta: int | None = None
    message: str | None = None


def _get_web3(endpoint_or_provider: str | Web3) -> Web3:
    if isinstance(endpoint_or_provider, Web3):
        return endpoint_or_provider
    return Web3(HTTPProvider(endpoint_or_provider, request_kwargs={'timeout': 30}))


def _mine_gas_price(
    nonce: int,
    gas_amount: int,
    from_address: str,
    difficulty: int = DEFAULT_DIFFICULTY,
    timeout_seconds: int = DEFAULT_MINE_TIMEOUT_SECONDS,
) -> tuple[int, int, float]:
    if difficulty < 1:
        raise ValueError('difficulty must be >= 1')

    nonce_hash = int.from_bytes(
        Web3.keccak(nonce.to_bytes(32, byteorder='big', signed=False)),
        byteorder='big',
        signed=False,
    )
    address_hash = int.from_bytes(
        Web3.keccak(hexstr=Web3.to_checksum_address(from_address)),
        byteorder='big',
        signed=False,
    )
    seed = nonce_hash ^ address_hash
    div_constant = MAX_NUMBER // difficulty

    start = perf_counter()
    iterations = 0

    while True:
        candidate = secrets.token_bytes(32)
        candidate_hash = int.from_bytes(Web3.keccak(candidate), byteorder='big', signed=False)
        result_hash = seed ^ candidate_hash
        if result_hash != 0 and div_constant // result_hash >= gas_amount:
            elapsed_ms = (perf_counter() - start) * 1_000.0
            gas_price = int.from_bytes(candidate, byteorder='big', signed=False)
            return gas_price, iterations + 1, elapsed_ms
        iterations += 1
        if timeout_seconds > 0 and (perf_counter() - start) >= timeout_seconds:
            raise TimeoutError(f'PoW mining exceeded timeout ({timeout_seconds}s).')


def mine_sfuel(
    endpoint_or_provider: str | Web3,
    beneficiary_address: str,
    faucet_address: str | None = None,
    faucet_func: str | None = None,
) -> MiningResult:
    web3 = _get_web3(endpoint_or_provider)
    if not web3.is_connected():
        raise ConnectionError('Unable to connect to provider.')

    chain_id = int(web3.eth.chain_id)

    if faucet_address and faucet_func:
        function_selector = faucet_func
    else:
        faucet = FAUCET_BY_CHAIN_ID.get(chain_id)
        if faucet is None:
            return MiningResult(
                success=False,
                message=f'Unsupported chain id for sFUEL faucet: {chain_id}',
            )
        faucet_address, function_selector = faucet

    if len(function_selector) != 10 or not function_selector.startswith('0x'):
        return MiningResult(success=False, message='Invalid faucet function selector.')

    faucet_checksum = web3.to_checksum_address(faucet_address)
    if len(web3.eth.get_code(faucet_checksum)) == 0:
        return MiningResult(
            success=False,
            message=f'Configured faucet has no contract code: {faucet_checksum}',
        )

    beneficiary_checksum = web3.to_checksum_address(beneficiary_address)
    start_balance = int(web3.eth.get_balance(beneficiary_checksum))

    ephemeral = Account.create()
    nonce = int(web3.eth.get_transaction_count(ephemeral.address))
    gas_price, iterations, elapsed_ms = _mine_gas_price(
        nonce=nonce,
        gas_amount=DEFAULT_TARGET_GAS,
        from_address=ephemeral.address,
    )

    encoded_param = web3.codec.encode(['address'], [beneficiary_checksum]).hex()
    data = function_selector + encoded_param
    tx = {
        'nonce': nonce,
        'to': faucet_checksum,
        'value': 0,
        'gas': DEFAULT_TARGET_GAS,
        'gasPrice': gas_price,
        'chainId': chain_id,
        'data': data,
    }

    signed = web3.eth.account.sign_transaction(tx, private_key=ephemeral.key)
    tx_hash = web3.eth.send_raw_transaction(signed.raw_transaction)
    receipt = web3.eth.wait_for_transaction_receipt(
        tx_hash,
        timeout=DEFAULT_RECEIPT_TIMEOUT_SECONDS,
    )

    receipt_status = int(receipt.get('status', 0))
    end_balance = int(web3.eth.get_balance(beneficiary_checksum))
    balance_delta = end_balance - start_balance
    success = receipt_status == 1 and balance_delta > 0

    if success:
        message = 'sFUEL funding completed.'
    elif receipt_status != 1:
        message = f'Funding transaction failed with receipt status {receipt_status}.'
    else:
        message = 'Funding transaction succeeded but beneficiary balance did not increase.'

    return MiningResult(
        success=success,
        tx_hash=tx_hash,
        receipt_status=receipt_status,
        gas_price=gas_price,
        iterations=iterations,
        elapsed_ms=elapsed_ms,
        start_balance=start_balance,
        end_balance=end_balance,
        balance_delta=balance_delta,
        message=message,
    )



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2021 SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from skale.types.schain import SchainOption
from enum import Enum

from eth_typing import HexStr
from web3 import Web3


class AllocationType(int, Enum):
    DEFAULT = 0
    NO_FILESTORAGE = 1
    MAX_CONTRACT_STORAGE = 2
    MAX_CONSENSUS_DB = 3
    MAX_FILESTORAGE = 4


@dataclass
class SchainOptions:
    multitransaction_mode: bool
    threshold_encryption: bool
    allocation_type: AllocationType
    external_gas_difficulty: HexStr
    min_gas_price: int | None
    max_gas_price: int | None

    def to_tuples(self) -> list[SchainOption]:
        options = [
            ('multitr', bool_to_bytes(self.multitransaction_mode)),
            ('encrypt', bool_to_bytes(self.threshold_encryption)),
            ('alloc', int_to_bytes(self.allocation_type.value)),
            ('powdifficulty', hex_str_to_bytes(self.external_gas_difficulty)),
        ]
        if self.min_gas_price is not None:
            options.append(('mingasprice', int_to_bytes(self.min_gas_price)))
        if self.max_gas_price is not None:
            options.append(('maxgasprice', int_to_bytes(self.max_gas_price)))
        return options


def parse_schain_options(raw_options: list[SchainOption]) -> SchainOptions:
    """
    Parses raw sChain options from smart contracts (list of tuples).
    Returns default values if nothing is set on contracts.
    """
    options_map = {k: v for k, v in raw_options}
    default_options = get_default_schain_options()

    def get_val(key, converter, default):
        return converter(options_map[key]) if key in options_map else default

    return SchainOptions(
        multitransaction_mode=get_val(
            'multitr', bytes_to_bool, default_options.multitransaction_mode
        ),
        threshold_encryption=get_val(
            'encrypt', bytes_to_bool, default_options.threshold_encryption
        ),
        allocation_type=get_val(
            'alloc',
            lambda x: AllocationType(bytes_to_int(x)),
            default_options.allocation_type,
        ),
        external_gas_difficulty=get_val(
            'powdifficulty', bytes_to_hex_str, default_options.external_gas_difficulty
        ),
        min_gas_price=get_val('mingasprice', bytes_to_int, default_options.min_gas_price),
        max_gas_price=get_val('maxgasprice', bytes_to_int, default_options.max_gas_price),
    )


def get_default_schain_options() -> SchainOptions:
    return SchainOptions(
        multitransaction_mode=False,
        threshold_encryption=False,
        allocation_type=AllocationType.DEFAULT,
        external_gas_difficulty=HexStr('0x01'),
        min_gas_price=None,
        max_gas_price=None,
    )


def bool_to_bytes(bool_value: bool) -> bytes:
    return bool_value.to_bytes(1, byteorder='big')


def int_to_bytes(int_value: int) -> bytes:
    return int_value.to_bytes(max(1, (int_value.bit_length() + 7) // 8), byteorder='big')


def bytes_to_int(bytes_value: bytes) -> int:
    return int.from_bytes(bytes_value, byteorder='big')


def bytes_to_bool(bytes_value: bytes) -> bool:
    return bool(int.from_bytes(bytes_value, 'big'))


def hex_str_to_bytes(hex_str: HexStr) -> bytes:
    return Web3.to_bytes(hexstr=hex_str)


def bytes_to_hex_str(bytes_value: bytes) -> HexStr:
    return Web3.to_hex(bytes_value)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from time import sleep

from eth_typing import ChecksumAddress
from web3.types import Wei

from skale.contracts.skale_contract import SkaleContract


class EthErc20(SkaleContract):
    def balance_of(self, address: ChecksumAddress) -> Wei:
        return Wei(self.contract.functions.balanceOf(address).call())

    def wait_for_balance_change(
        self,
        address: ChecksumAddress,
        initial_balance: Wei,
        timeout: int = 120,
        poll_interval: int = 5,
        raise_on_timeout: bool = False,
    ) -> bool:
        iterations = timeout // poll_interval + 1
        for _ in range(iterations):
            current_balance = self.balance_of(address)
            if current_balance != initial_balance:
                return True
            sleep(poll_interval)

        if raise_on_timeout:
            raise TimeoutError(f'Balance did not change for {address} within {timeout} seconds')
        return False



"""SKALE Allocator test"""

from functools import cached_property

from skale import SkaleAllocator
from skale.utils.helper import get_allocator_address
from skale.utils.web3_utils import init_web3
from skale.wallets import Web3Wallet
from tests.constants import ENDPOINT, ETH_PRIVATE_KEY, TEST_ALLOCATOR_ABI_FILEPATH

ALLOCATOR_CONTRACTS_NUMBER = 2


def test_init_allocator():
    web3 = init_web3(ENDPOINT)
    wallet = Web3Wallet(ETH_PRIVATE_KEY, web3)
    skale_allocator = SkaleAllocator(
        ENDPOINT, get_allocator_address(TEST_ALLOCATOR_ABI_FILEPATH), wallet, provider_timeout=20
    )
    cached_properties = [
        attr
        for attr in dir(skale_allocator)
        if isinstance(getattr(type(skale_allocator), attr, None), cached_property)
    ]

    assert len(cached_properties) == ALLOCATOR_CONTRACTS_NUMBER
    assert skale_allocator.allocator
    assert skale_allocator.escrow



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2024-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum
from typing import NewType, TypedDict

from eth_typing import ChecksumAddress
from web3.types import Wei

from skale.types.validator import ValidatorId

DelegationId = NewType('DelegationId', int)


class DelegationStatus(Enum):
    PROPOSED = 0
    ACCEPTED = 1
    CANCELED = 2
    REJECTED = 3
    DELEGATED = 4
    UNDELEGATION_REQUESTED = 5
    COMPLETED = 6


class Delegation(TypedDict):
    address: ChecksumAddress
    validator_id: ValidatorId
    amount: Wei
    delegation_period: int
    created: int
    started: int
    finished: int
    info: str


class FullDelegation(Delegation):
    id: DelegationId
    status: DelegationStatus



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
"""NodeRotation.sol functions"""

from __future__ import annotations

import functools
import logging
from typing import TYPE_CHECKING, List

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction
from web3.exceptions import ContractLogicError

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.types.node import NodeId
from skale.types.rotation import Rotation, RotationSwap
from skale.types.schain import SchainHash, SchainName
from skale.utils.helper import is_test_env, schain_name_to_hash

if TYPE_CHECKING:
    from skale.contracts.manager.schains import SChains


logger = logging.getLogger(__name__)


NO_PREVIOUS_NODE_EXCEPTION_TEXT = 'No previous node'


class NodeRotation(SkaleManagerContract):
    """Wrapper for NodeRotation.sol functions"""

    @property
    @functools.lru_cache()
    def schains(self) -> SChains:
        return self.skale.schains

    def get_rotation(self, schain_name: SchainName) -> Rotation:
        schain_hash = schain_name_to_hash(schain_name)
        rotation_data = self.contract.functions.getRotation(schain_hash).call()
        return Rotation(*rotation_data)

    def get_leaving_history(self, node_id: NodeId) -> List[RotationSwap]:
        raw_history = self.contract.functions.getLeavingHistory(node_id).call()
        history = [
            RotationSwap({'schain_id': SchainHash(schain[0]), 'finished_rotation': int(schain[1])})
            for schain in raw_history
        ]
        return history

    def get_schain_finish_ts(self, node_id: NodeId, schain_name: SchainName) -> int | None:
        history = self.get_leaving_history(node_id)
        schain_hash = schain_name_to_hash(schain_name)
        finish_ts = next(
            (swap['finished_rotation'] for swap in history if swap['schain_id'] == schain_hash),
            None,
        )
        if not finish_ts:
            return None
        return int(finish_ts)

    def is_rotation_in_progress(self, schain_name: SchainName) -> bool:
        schain_hash = schain_name_to_hash(schain_name)
        return bool(self.contract.functions.isRotationInProgress(schain_hash).call())

    def is_new_node_found(self, schain_name: SchainName) -> bool:
        schain_hash = schain_name_to_hash(schain_name)
        return bool(self.contract.functions.isNewNodeFound(schain_hash).call())

    def is_rotation_active(self, schain_name: SchainName) -> bool:
        """
        The public function that tells whether rotation is in the active phase - the new group is
        already generated
        """
        finish_ts_reached = self.is_finish_ts_reached(schain_name)
        return self.is_rotation_in_progress(schain_name) and not finish_ts_reached

    def is_finish_ts_reached(self, schain_name: SchainName) -> bool:
        rotation = self.skale.node_rotation.get_rotation(schain_name)
        schain_finish_ts = self.get_schain_finish_ts(rotation.leaving_node_id, schain_name)

        if not schain_finish_ts:
            schain_finish_ts = 0

        latest_block = self.skale.web3.eth.get_block('latest')
        current_ts = latest_block['timestamp']

        logger.info(f'current_ts: {current_ts}, schain_finish_ts: {schain_finish_ts}')
        return current_ts > schain_finish_ts

    def wait_for_new_node(self, schain_name: SchainName) -> bool:
        schain_hash = schain_name_to_hash(schain_name)
        return bool(self.contract.functions.waitForNewNode(schain_hash).call())

    @transaction_method
    def grant_role(self, role: bytes, owner: str) -> 'ContractFunction':
        return self.contract.functions.grantRole(role, owner)

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.hasRole(role, address).call())

    def debugger_role(self) -> bytes:
        return bytes(self.contract.functions.DEBUGGER_ROLE().call())

    def get_previous_node(self, schain_name: SchainName, node_id: NodeId) -> NodeId | None:
        schain_hash = schain_name_to_hash(schain_name)
        try:
            return NodeId(self.contract.functions.getPreviousNode(schain_hash, node_id).call())
        except (ContractLogicError, ValueError) as e:
            if NO_PREVIOUS_NODE_EXCEPTION_TEXT in str(e):
                return None
            if is_test_env():
                logger.debug(f'Error in test environment: {e}')
                return None
            raise e



def test_slashing_table_get_set(skale):
    penalty_value = 50
    offense = 'FailedDkg'
    skale.slashing_table.set_penalty(offense, penalty_value, wait_for=True)

    expected_penalty = penalty_value
    penalty = skale.slashing_table.get_penalty(offense)
    assert penalty == expected_penalty



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.fair_manager import FairManager
from skale.types.committee import CommitteeIndex
from skale.types.node import FairNode, FairNodeForChainConfig
from skale.utils.constants import ZERO_ADDRESS


def convert_to_node_for_chain_config(fair: FairManager, node: FairNode) -> FairNodeForChainConfig:
    reward_wallet_address = fair.web3.to_checksum_address(ZERO_ADDRESS)
    # Checking if the node is in boot node group i.e. is in the initial committee
    initial_committee_index: CommitteeIndex = CommitteeIndex(0)
    initial_committee = fair.committee.get_committee(initial_committee_index)
    public_key = fair.nodes.get_public_key(node.id)
    if node.id not in initial_committee.node_ids:
        reward_wallet_address = fair.staking.get_reward_wallet(node.id)
    return FairNodeForChainConfig(
        **node.to_dict(),
        ip=node.ip,
        reward_wallet_address=reward_wallet_address,
        public_key=public_key,
    )



import tomllib
from pathlib import Path

import pytest
from pydantic import ValidationError
from skale_core.settings import (
    BaseNodeSettings,
    FairBaseSettings,
    FairSettings,
    SkalePassiveSettings,
    SkaleSettings,
    _resolve_type,
    write_internal_settings_file,
    write_node_settings_file,
)
from skale_core.types import NodeMode, NodeType

BASE_NODE_DATA = {
    'env_type': 'mainnet',
    'endpoint': 'http://localhost:8545',
    'node_version': '1.0.0',
    'block_device': '/dev/sda',
}

SGX_URL = 'https://sgx.example.com'


@pytest.fixture(
    params=[
        ('skale', 'active'),
        ('skale', 'passive'),
        ('fair', 'active'),
        ('fair', 'passive'),
    ]
)
def node_type_mode(request: pytest.FixtureRequest) -> tuple[NodeType, NodeMode]:
    return request.param


def test_write_internal_settings(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    node_type_mode: tuple[NodeType, NodeMode],
) -> None:
    node_type, node_mode = node_type_mode
    monkeypatch.setenv('SKALE_DIR_HOST', str(tmp_path))
    path = tmp_path / 'node.toml'
    data = {'node_type': node_type, 'node_mode': node_mode, 'skale_dir_host': str(tmp_path)}
    result = write_internal_settings_file(path=path, data=data)
    assert result.node_type == node_type
    assert result.node_mode == node_mode
    with open(path, 'rb') as f:
        data = tomllib.load(f)
    assert data['node_type'] == node_type
    assert data['node_mode'] == node_mode


def test_write_internal_settings_invalid_type(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv('SKALE_DIR_HOST', str(tmp_path))
    with pytest.raises(ValidationError):
        write_internal_settings_file(
            path=tmp_path / 'node.toml',
            data={'node_type': 'unknown', 'node_mode': 'active', 'skale_dir_host': str(tmp_path)},
        )


@pytest.fixture(
    params=[
        (
            SkaleSettings,
            {
                **BASE_NODE_DATA,
                'manager_contracts': '0x1',
                'ima_contracts': '0x2',
                'sgx_url': SGX_URL,
                'docker_lvmpy_version': '1.0.0',
            },
        ),
        (
            SkalePassiveSettings,
            {
                **BASE_NODE_DATA,
                'manager_contracts': '0x1',
                'ima_contracts': '0x2',
                'schain_name': 'test-chain',
            },
        ),
        (FairBaseSettings, {**BASE_NODE_DATA, 'fair_contracts': '0x3'}),
        (
            FairSettings,
            {**BASE_NODE_DATA, 'fair_contracts': '0x3', 'sgx_url': SGX_URL},
        ),
    ]
)
def node_settings_entry(
    request: pytest.FixtureRequest,
) -> tuple[type[BaseNodeSettings], dict]:
    return request.param


def test_write_node_settings(
    tmp_path: Path, node_settings_entry: tuple[type[BaseNodeSettings], dict]
) -> None:
    settings_type, data = node_settings_entry
    path = tmp_path / 'settings.toml'
    result = write_node_settings_file(path=path, settings_type=settings_type, data=data)
    assert isinstance(result, settings_type)
    assert str(result.endpoint) == data['endpoint'] + '/'
    assert path.exists()
    with open(path, 'rb') as f:
        written = tomllib.load(f)
    assert written['env_type'] == data['env_type']


def test_node_settings_rejects_missing_fields(tmp_path: Path) -> None:
    with pytest.raises(ValidationError):
        write_node_settings_file(
            path=tmp_path / 'settings.toml',
            settings_type=SkaleSettings,
            data={'env_type': 'mainnet'},
        )


def test_resolve_type() -> None:
    assert _resolve_type('skale', 'active') is SkaleSettings
    assert _resolve_type('skale', 'passive') is SkalePassiveSettings
    assert _resolve_type('fair', 'active') is FairSettings
    assert _resolve_type('fair', 'passive') is FairBaseSettings
    assert _resolve_type('skale', 'unknown') is BaseNodeSettings



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from functools import cached_property

from skale_contracts.project_factory import SkaleProject
from skale_contracts.projects.skale_manager import SkaleManagerContract

from skale.contracts.manager.bounty_v2 import BountyV2
from skale.contracts.manager.constants_holder import ConstantsHolder
from skale.contracts.manager.contract_manager import ContractManager
from skale.contracts.manager.delegation.delegation_controller import DelegationController
from skale.contracts.manager.delegation.delegation_period_manager import DelegationPeriodManager
from skale.contracts.manager.delegation.distributor import Distributor
from skale.contracts.manager.delegation.slashing_table import SlashingTable
from skale.contracts.manager.delegation.token_state import TokenState
from skale.contracts.manager.delegation.validator_service import ValidatorService
from skale.contracts.manager.dkg import DKG
from skale.contracts.manager.key_storage import KeyStorage
from skale.contracts.manager.manager import Manager
from skale.contracts.manager.node_rotation import NodeRotation
from skale.contracts.manager.nodes import Nodes
from skale.contracts.manager.punisher import Punisher
from skale.contracts.manager.schains import SChains
from skale.contracts.manager.schains_internal import SChainsInternal
from skale.contracts.manager.sync_manager import SyncManager
from skale.contracts.manager.test.time_helpers_with_debug import TimeHelpersWithDebug
from skale.contracts.manager.token import Token
from skale.contracts.manager.wallets import Wallets
from skale.skale_base import SkaleBase


class SkaleManager(SkaleBase):
    """Represents skale-manager smart contracts"""

    @property
    def project_name(self) -> SkaleProject:
        return SkaleProject.SKALE_MANAGER

    @cached_property
    def contract_manager(self) -> ContractManager:
        return ContractManager(self, SkaleManagerContract.CONTRACT_MANAGER)

    @cached_property
    def token(self) -> Token:
        return Token(self, SkaleManagerContract.SKALE_TOKEN)

    @cached_property
    def manager(self) -> Manager:
        return Manager(self, SkaleManagerContract.SKALE_MANAGER)

    @cached_property
    def constants_holder(self) -> ConstantsHolder:
        return ConstantsHolder(self, SkaleManagerContract.CONSTANTS_HOLDER)

    @cached_property
    def nodes(self) -> Nodes:
        return Nodes(self, SkaleManagerContract.NODES)

    @cached_property
    def node_rotation(self) -> NodeRotation:
        return NodeRotation(self, SkaleManagerContract.NODE_ROTATION)

    @cached_property
    def schains(self) -> SChains:
        return SChains(self, SkaleManagerContract.SCHAINS)

    @cached_property
    def schains_internal(self) -> SChainsInternal:
        return SChainsInternal(self, SkaleManagerContract.SCHAINS_INTERNAL)

    @cached_property
    def dkg(self) -> DKG:
        return DKG(self, SkaleManagerContract.SKALE_DKG)

    @cached_property
    def key_storage(self) -> KeyStorage:
        return KeyStorage(self, SkaleManagerContract.KEY_STORAGE)

    @cached_property
    def delegation_controller(self) -> DelegationController:
        return DelegationController(self, SkaleManagerContract.DELEGATION_CONTROLLER)

    @cached_property
    def delegation_period_manager(self) -> DelegationPeriodManager:
        return DelegationPeriodManager(self, SkaleManagerContract.DELEGATION_PERIOD_MANAGER)

    @cached_property
    def validator_service(self) -> ValidatorService:
        return ValidatorService(self, SkaleManagerContract.VALIDATOR_SERVICE)

    @cached_property
    def token_state(self) -> TokenState:
        return TokenState(self, SkaleManagerContract.TOKEN_STATE)

    @cached_property
    def distributor(self) -> Distributor:
        return Distributor(self, SkaleManagerContract.DISTRIBUTOR)

    @cached_property
    def slashing_table(self) -> SlashingTable:
        return SlashingTable(self, SkaleManagerContract.SLASHING_TABLE)

    @cached_property
    def wallets(self) -> Wallets:
        return Wallets(self, SkaleManagerContract.WALLETS)

    @cached_property
    def bounty_v2(self) -> BountyV2:
        return BountyV2(self, SkaleManagerContract.BOUNTY_V2)

    @cached_property
    def punisher(self) -> Punisher:
        return Punisher(self, SkaleManagerContract.PUNISHER)

    @cached_property
    def sync_manager(self) -> SyncManager:
        return SyncManager(self, SkaleManagerContract.SYNC_MANAGER)

    @cached_property
    def time_helpers_with_debug(self) -> TimeHelpersWithDebug:
        return TimeHelpersWithDebug(self, SkaleManagerContract.TIME_HELPERS_WITH_DEBUG)


def spawn_skale_manager_lib(skale: SkaleManager) -> SkaleManager:
    return SkaleManager(skale._endpoint, skale.instance.address, skale.wallet)



"""SKALE main test"""

from functools import cached_property

import pytest
from skale_contracts.projects.skale_manager import SkaleManagerContract
from web3 import HTTPProvider, LegacyWebSocketProvider, Web3

from skale import SkaleManager
from skale.contracts.manager.nodes import Nodes
from skale.contracts.skale_contract import SkaleContract
from skale.utils.helper import get_skale_manager_address
from skale.utils.web3_utils import init_web3
from skale.wallets import Web3Wallet
from tests.constants import ENDPOINT, ETH_PRIVATE_KEY, TEST_ABI_FILEPATH

ALLOWED_SKIP_TIME_GAP = 10
MANAGER_CONTRACTS_NUMBER = 21


def test_lib_init():
    web3 = init_web3(ENDPOINT)
    wallet = Web3Wallet(ETH_PRIVATE_KEY, web3)
    skale = SkaleManager(
        ENDPOINT, get_skale_manager_address(TEST_ABI_FILEPATH), wallet, provider_timeout=20
    )

    cached_properties = [
        attr for attr in dir(skale) if isinstance(getattr(type(skale), attr, None), cached_property)
    ]

    assert len(cached_properties) == MANAGER_CONTRACTS_NUMBER

    for prop in cached_properties:
        value = getattr(skale, prop)
        assert value is not None, f'Cached property {prop} returned None'
        assert issubclass(type(value), SkaleContract), (
            f'Cached property {prop} is not a subclass of SkaleContract'
        )

    isinstance(skale.web3.provider, HTTPProvider)

    ws_endpoint = 'ws://localhost:8545'

    skale = SkaleManager(ws_endpoint, get_skale_manager_address(TEST_ABI_FILEPATH), wallet)
    assert skale.web3.provider.websocket_timeout == 30
    assert skale.web3.provider.conn.websocket_kwargs == {'max_size': 5 * 1024 * 1024}
    assert isinstance(skale.web3.provider, LegacyWebSocketProvider)

    file_endpoint = 'file://local_file:1001'
    with pytest.raises(Exception):
        SkaleManager(file_endpoint, get_skale_manager_address(TEST_ABI_FILEPATH), wallet)


def test_contract_cache():
    web3 = init_web3(ENDPOINT)
    wallet = Web3Wallet(ETH_PRIVATE_KEY, web3)
    skale = SkaleManager(
        ENDPOINT, get_skale_manager_address(TEST_ABI_FILEPATH), wallet, provider_timeout=20
    )
    first_access = skale.nodes
    second_access = skale.nodes
    assert first_access is second_access


def test_get_contract_address(skale):
    lib_nodes_address = skale.instance.get_contract_address(SkaleManagerContract.NODES)
    lib_nodes_address = Web3.to_checksum_address(lib_nodes_address)
    nodes_address = skale.nodes.address

    assert lib_nodes_address == nodes_address


def test_get_attr(skale):
    with pytest.raises(AttributeError):
        skale.t123_random_attr
    skale_py_nodes_contract = skale.nodes
    assert issubclass(type(skale_py_nodes_contract), SkaleContract)
    assert isinstance(skale_py_nodes_contract, Nodes)


def test_legacy_abi_generation(skale):
    abi_dict = skale._generate_legacy_abi()
    assert 'nodes_abi' in abi_dict
    assert 'nodes_address' in abi_dict
    assert isinstance(abi_dict['nodes_abi'], list)
    assert Web3.is_checksum_address(abi_dict['nodes_address'])



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


from typing import List, Tuple

from eth_abi import encode
from eth_typing import ChecksumAddress
from eth_utils import function_signature_to_4byte_selector
from multisigwallet_predeployed import MULTISIGWALLET_ADDRESS, MultiSigWalletGenerator
from skale_contracts.projects.marionette import MarionetteInstance
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.types import TxParams, Wei

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.wallets import BaseWallet


class MultiSigContract(BaseContract):
    def __init__(
        self,
        web3: Web3,
        wallet: BaseWallet | None = None,
    ):
        multisigwallet_generator = MultiSigWalletGenerator()
        abi = multisigwallet_generator.get_abi()
        multisig_address = Web3.to_checksum_address(MULTISIGWALLET_ADDRESS)
        super().__init__(web3, multisig_address, abi, wallet)

    def confirmations(self, transaction_id: int, owner_address: ChecksumAddress) -> bool:
        return self.contract.functions.confirmations(transaction_id, owner_address).call()

    def get_confirmation_count(self, transaction_id: int) -> int:
        return self.contract.functions.getConfirmationCount(transaction_id).call()

    def get_confirmations(self, transaction_id: int) -> List[ChecksumAddress]:
        return self.contract.functions.getConfirmations(transaction_id).call()

    def get_owners(self) -> List[ChecksumAddress]:
        return self.contract.functions.getOwners().call()

    def get_transaction_count(self, pending: bool, executed: bool) -> int:
        return self.contract.functions.getTransactionCount(pending, executed).call()

    def get_transaction_ids(
        self, from_tx: int, to_tx: int, pending: bool, executed: bool
    ) -> List[int]:
        return self.contract.functions.getTransactionIds(from_tx, to_tx, pending, executed).call()

    def is_confirmed(self, transaction_id: int) -> bool:
        return self.contract.functions.isConfirmed(transaction_id).call()

    def is_owner(self, address: ChecksumAddress) -> bool:
        return self.contract.functions.isOwner(address).call()

    def owners(self, index: int) -> ChecksumAddress:
        return self.contract.functions.owners(index).call()

    def required(self) -> int:
        return self.contract.functions.required().call()

    def transaction_count(self) -> int:
        return self.contract.functions.transactionCount().call()

    def transactions(self, index: int) -> Tuple[ChecksumAddress, int, bytes, bool]:
        return self.contract.functions.transactions(index).call()

    @transaction_method
    def add_owner(self, owner: ChecksumAddress) -> ContractFunction:
        func = self.contract.functions.addOwner(owner)
        return self._submit_self_transaction(func)

    @transaction_method
    def change_requirement(self, required: int) -> ContractFunction:
        func = self.contract.functions.changeRequirement(required)
        return self._submit_self_transaction(func)

    @transaction_method
    def remove_owner(self, owner: ChecksumAddress) -> ContractFunction:
        func = self.contract.functions.removeOwner(owner)
        return self._submit_self_transaction(func)

    @transaction_method
    def replace_owner(self, owner: ChecksumAddress, new_owner: ChecksumAddress) -> ContractFunction:
        func = self.contract.functions.replaceOwner(owner, new_owner)
        return self._submit_self_transaction(func)

    @transaction_method
    def confirm_transaction(self, transaction_id: int) -> ContractFunction:
        return self.contract.functions.confirmTransaction(transaction_id)

    @transaction_method
    def execute_transaction(self, transaction_id: int) -> ContractFunction:
        return self.contract.functions.executeTransaction(transaction_id)

    @transaction_method
    def revoke_confirmation(self, transaction_id: int) -> ContractFunction:
        return self.contract.functions.revokeConfirmation(transaction_id)

    def submit_transaction(
        self,
        destination: ChecksumAddress,
        value: int,
        data: bytes,
    ) -> ContractFunction:
        return self.contract.functions.submitTransaction(
            destination,
            value,
            data,
        )

    def _encode_transaction_data(self, contract_function: ContractFunction) -> bytes:
        tx_params: TxParams = {'gas': Wei(0), 'gasPrice': Wei(0)}
        tx = contract_function.build_transaction(tx_params)
        data = tx['data']
        if isinstance(data, str):
            return bytes.fromhex(data[2:])
        return data

    def _submit_self_transaction(self, contract_function: ContractFunction) -> ContractFunction:
        calldata = self._encode_transaction_data(contract_function)
        return self.submit_transaction(self.address, 0, calldata)

    def _encode_marionette_execute(self, destination: ChecksumAddress, data: bytes) -> bytes:
        function_selector = function_signature_to_4byte_selector('execute(address,uint256,bytes)')

        encoded_params = encode(['address', 'uint256', 'bytes'], [destination, 0, data])
        return function_selector + encoded_params

    @transaction_method
    def execute_via_marionette(
        self,
        destination: ChecksumAddress,
        contract_function: ContractFunction,
    ) -> ContractFunction:
        marionette_address = MarionetteInstance.PREDEPLOYED['Marionette']
        target_calldata = self._encode_transaction_data(contract_function)

        marionette_calldata = self._encode_marionette_execute(destination, target_calldata)
        return self.submit_transaction(marionette_address, 0, marionette_calldata)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2021-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, List, TypedDict

from skale.types.rotation import RotationNodeData
from skale.utils.helper import schain_name_to_hash

if TYPE_CHECKING:
    from skale.skale_manager import SkaleManager
    from skale.types.dkg import G2Point
    from skale.types.node import NodeId
    from skale.types.rotation import BlsPublicKey, NodeGroups, Rotation
    from skale.types.schain import SchainName

logger = logging.getLogger(__name__)


class PreviousNodeData(TypedDict):
    finish_ts: int
    previous_node_id: NodeId


def get_previous_schain_groups(
    skale: SkaleManager,
    schain_name: SchainName,
    leaving_node_id: NodeId | None = None,
) -> NodeGroups:
    """
    Returns all previous node groups with public keys and finish timestamps.
    In case of no rotations returns the current state.
    """
    logger.info(f'Collecting rotation history for {schain_name}...')
    node_groups: NodeGroups = {}

    group_id = schain_name_to_hash(schain_name)

    previous_public_keys = skale.key_storage.get_all_previous_public_keys(group_id)
    current_public_key = skale.key_storage.get_common_public_key(group_id)

    rotation = skale.node_rotation.get_rotation(schain_name)

    logger.info(f'Rotation data for {schain_name}: {rotation}')

    _add_current_schain_state(skale, node_groups, rotation, schain_name, current_public_key)
    if rotation.rotation_counter == 0:
        return node_groups

    _add_previous_schain_rotations_state(
        skale=skale,
        node_groups=node_groups,
        rotation=rotation,
        schain_name=schain_name,
        previous_public_keys=previous_public_keys,
        leaving_node_id=leaving_node_id,
    )

    return node_groups


def _add_current_schain_state(
    skale: SkaleManager,
    node_groups: NodeGroups,
    rotation: Rotation,
    schain_name: SchainName,
    current_public_key: G2Point,
) -> None:
    """
    Internal function, composes the initial info about the current sChain state and adds it to the
    node_groups dictionary
    """
    current_nodes = {}
    ids = skale.schains_internal.node_ids_for_schain(schain_name)
    for index, node_id in enumerate(ids):
        public_key = skale.nodes.node_public_key(node_id)
        current_nodes[node_id] = RotationNodeData(index, node_id, public_key)

    node_groups[rotation.rotation_counter] = {
        'rotation': None,
        'nodes': current_nodes,
        'finish_ts': None,
        'bls_public_key': _compose_bls_public_key_info(current_public_key),
    }


def _add_previous_schain_rotations_state(
    skale: SkaleManager,
    node_groups: NodeGroups,
    rotation: Rotation,
    schain_name: SchainName,
    previous_public_keys: list[G2Point],
    leaving_node_id: NodeId | None = None,
) -> None:
    """
    Internal function, handles rotations from (rotation_counter - 2) to 0 and adds them to the
    node_groups dictionary
    """

    previous_nodes: Dict[NodeId, PreviousNodeData] = {}

    for rotation_id in range(rotation.rotation_counter - 1, -1, -1):
        nodes = node_groups[rotation_id + 1]['nodes'].copy()
        for node_id in nodes:
            if node_id not in previous_nodes:
                previous_node = skale.node_rotation.get_previous_node(schain_name, node_id)
                if previous_node is not None:
                    finish_ts = skale.node_rotation.get_schain_finish_ts(previous_node, schain_name)
                    previous_nodes[node_id] = {
                        'finish_ts': finish_ts or 0,
                        'previous_node_id': previous_node,
                    }

        new_node_id = max(previous_nodes.items(), key=lambda x: x[1]['finish_ts'])[0]
        previous_node_id = previous_nodes[new_node_id]['previous_node_id']
        public_key = skale.nodes.node_public_key(previous_node_id)

        current_finish_ts = previous_nodes[new_node_id]['finish_ts']
        next_dkg_is_failed = current_finish_ts + 1 == node_groups[rotation_id + 1]['finish_ts']

        nodes[previous_node_id] = RotationNodeData(
            nodes[new_node_id].index, previous_node_id, public_key
        )
        del nodes[new_node_id]

        bls_public_key = None
        if not next_dkg_is_failed:
            if len(previous_public_keys) > 0:
                bls_public_key = _pop_previous_bls_public_key(previous_public_keys)
        else:
            bls_public_key = node_groups[rotation_id + 1]['bls_public_key']
            node_groups[rotation_id + 1]['finish_ts'] = None
            node_groups[rotation_id + 1]['bls_public_key'] = None

        logger.info(f'Adding rotation: {previous_node_id} -> {new_node_id}')

        node_groups[rotation_id] = {
            'rotation': {'leaving_node_id': previous_node_id, 'new_node_id': new_node_id},
            'nodes': nodes,
            'finish_ts': current_finish_ts,
            'bls_public_key': bls_public_key,
        }

        del previous_nodes[new_node_id]

        if leaving_node_id and previous_node_id == leaving_node_id:
            logger.info(f'Finishing rotation history parsing: {leaving_node_id} found')
            break


def _pop_previous_bls_public_key(previous_public_keys: List[G2Point]) -> BlsPublicKey | None:
    """
    Returns BLS public key for the group and removes it from the list, returns None if node
    with provided node_id was kicked out of the chain because of failed DKG.
    """
    raw_bls_keys = previous_public_keys[-1]
    bls_keys = _compose_bls_public_key_info(raw_bls_keys)
    del previous_public_keys[-1]
    return bls_keys


def _compose_bls_public_key_info(bls_public_key: G2Point) -> BlsPublicKey | None:
    if bls_public_key:
        return {
            'blsPublicKey0': str(bls_public_key[0][0]),
            'blsPublicKey1': str(bls_public_key[0][1]),
            'blsPublicKey2': str(bls_public_key[1][0]),
            'blsPublicKey3': str(bls_public_key[1][1]),
        }
    return None



"""SKALE manager test"""

import random
from unittest import mock
from unittest.mock import Mock

import pytest
import web3
from hexbytes import HexBytes

from skale.skale_manager import SkaleManager
from skale.transactions.result import DryRunRevertError, TransactionFailedError
from skale.types.node import NodeId
from skale.utils.contracts_provision import DEFAULT_DOMAIN_NAME
from skale.utils.contracts_provision.main import (
    generate_random_node_data,
    generate_random_schain_data,
)
from skale.wallets import Web3Wallet
from skale.wallets.web3_wallet import generate_wallet
from tests.constants import TEST_GAS_LIMIT


def test_get_bounty(skale):
    node_id = 0
    nonce = skale.web3.eth.get_transaction_count(skale.wallet.address)
    contract_address = skale.manager.address
    chain_id = skale.web3.eth.chain_id
    expected_txn = {
        'value': 0,
        'gasPrice': skale.gas_price,
        'chainId': chain_id,
        'gas': TEST_GAS_LIMIT,
        'nonce': nonce,
        'type': 1,
        'to': contract_address,
        'data': ('0xee8c4bbf0000000000000000000000000000000000000000000000000000000000000000'),
    }
    exp = skale.web3.eth.account.sign_transaction(
        expected_txn, skale.wallet._private_key
    ).raw_transaction
    with mock.patch.object(
        skale.manager.contract.functions.getBounty, 'call', new=Mock(return_value=[])
    ):
        with mock.patch.object(web3.eth.Eth, 'send_raw_transaction') as send_tx_mock:
            send_tx_mock.return_value = b'hexstring'
            skale.manager.get_bounty(node_id, wait_for=False, gas_limit=TEST_GAS_LIMIT)
            send_tx_mock.assert_called_with(HexBytes(exp))


def test_create_delete_schain(skale, nodes):
    schains_ids = skale.schains_internal.all_schain_hashes()

    type_of_nodes, lifetime_seconds, name = generate_random_schain_data(skale)
    price_in_wei = skale.schains.schain_price(type_of_nodes, lifetime_seconds)

    try:
        tx_res = skale.manager.create_schain(
            lifetime_seconds, type_of_nodes, price_in_wei, name, wait_for=True
        )

        assert tx_res.receipt['status'] == 1

        schains_ids_number_after = skale.schains_internal.number_of_schains()
        assert schains_ids_number_after == len(schains_ids) + 1
        schains_ids_after = skale.schains_internal.all_schain_hashes()

        schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
        assert name in schains_names
    finally:
        skale.manager.delete_schain(name, wait_for=True)

    schains_ids_number_after = skale.schains_internal.number_of_schains()
    assert schains_ids_number_after == len(schains_ids)
    schains_ids_after = skale.schains_internal.all_schain_hashes()

    schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
    assert name not in schains_names


def test_delete_schain_by_root(skale, nodes):
    schains_ids = skale.schains_internal.all_schain_hashes()
    name = ''.join(random.choice('abcde') for _ in range(4))
    try:
        skale.manager.create_default_schain(name)
    finally:
        skale.manager.delete_schain_by_root(name, wait_for=True)

    schains_ids_number_after = skale.schains_internal.number_of_schains()
    assert schains_ids_number_after == len(schains_ids)
    schains_ids_after = skale.schains_internal.all_schain_hashes()

    schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
    assert name not in schains_names


def test_create_delete_default_schain(skale, nodes):
    schains_ids = skale.schains_internal.all_schain_hashes()
    _, _, name = generate_random_schain_data(skale)
    try:
        skale.manager.create_default_schain(name)

        schains_ids_number_after = skale.schains_internal.number_of_schains()
        assert schains_ids_number_after == len(schains_ids) + 1
        schains_ids_after = skale.schains_internal.all_schain_hashes()

        schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
        assert name in schains_names
    finally:
        skale.manager.delete_schain(name)

    schains_ids_number_after = skale.schains_internal.number_of_schains()
    assert schains_ids_number_after == len(schains_ids)
    schains_ids_after = skale.schains_internal.all_schain_hashes()

    schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
    assert name not in schains_names


def test_create_node_status_0(failed_skale):
    skale = failed_skale
    ip, public_ip, port, name = generate_random_node_data()
    tx_res = skale.manager.create_node(
        ip,
        port,
        name,
        domain_name=DEFAULT_DOMAIN_NAME,
        wait_for=True,
        skip_dry_run=True,
        raise_for_status=False,
    )
    assert tx_res.receipt['status'] == 0
    with pytest.raises(TransactionFailedError):
        tx_res.raise_for_status()


def test_node_exit_with_no_schains(skale: SkaleManager, node: tuple[NodeId, Web3Wallet]):
    node_id = node[0]
    skale.nodes.init_exit(node_id, wait_for=True)
    tx_res = skale.manager.node_exit(node_id, wait_for=True)
    assert tx_res.receipt['status'] == 1
    assert skale.nodes.node_status(node_id) == 2


def test_failed_node_exit(skale):
    not_existed_node_id = 1
    with pytest.raises(DryRunRevertError):
        skale.manager.node_exit(not_existed_node_id, wait_for=True, gas_limit=TEST_GAS_LIMIT)


def test_grant_role(skale):
    wallet = generate_wallet(skale.web3)
    default_admin_role = skale.manager.default_admin_role()
    assert not skale.manager.has_role(default_admin_role, wallet.address)

    skale.manager.grant_role(default_admin_role, wallet.address, wait_for=True)
    assert skale.manager.has_role(default_admin_role, wallet.address)


def test_grant_admin_role(skale):
    wallet = generate_wallet(skale.web3)
    admin_role = skale.manager.admin_role()
    assert not skale.manager.has_role(admin_role, wallet.address)

    skale.manager.grant_role(admin_role, wallet.address, wait_for=True)
    assert skale.manager.has_role(admin_role, wallet.address)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


import logging
import time
from functools import partial, wraps
from typing import Any, Callable, Optional

from eth_typing import ChecksumAddress
from web3 import Web3
from web3._utils.transactions import get_block_gas_limit
from web3.contract.contract import ContractFunction
from web3.exceptions import ContractLogicError, Web3Exception
from web3.types import Nonce, TxParams, Wei

import skale.config as config
from skale.transactions.exceptions import TransactionError
from skale.transactions.result import TxCallResult, TxRes, TxStatus
from skale.utils.web3_utils import get_eth_nonce
from skale.wallets import BaseWallet

logger = logging.getLogger(__name__)


DEFAULT_ETH_SEND_GAS_LIMIT = 22000


def make_dry_run_call(
    web3: Web3,
    wallet: BaseWallet,
    method: ContractFunction,
    gas_limit: int | None = None,
    value: Wei = Wei(0),
) -> TxCallResult:
    opts = TxParams({'from': wallet.address, 'value': value})
    logger.info(
        f'Dry run tx: {method.fn_name}, '
        f'sender: {wallet.address}, '
        f'wallet: {wallet.__class__.__name__}, '
        f'value: {value}, '
    )
    estimated_gas = 0

    try:
        if gas_limit:
            estimated_gas = gas_limit
            opts.update({'gas': gas_limit})
            method.call(opts)
        else:
            estimated_gas = estimate_gas(web3, method, opts)
        logger.info(f'Estimated gas for {method.fn_name}: {estimated_gas}')
    except ContractLogicError as e:
        message = e.message or 'Contract logic error'
        error_data = e.data or {}
        data = {'data': error_data} if isinstance(error_data, str) else error_data
        return TxCallResult(status=TxStatus.FAILED, error='revert', message=message, data=data)
    except (Web3Exception, ValueError) as e:
        logger.exception('Dry run for %s failed', method)
        return TxCallResult(status=TxStatus.FAILED, error='exception', message=str(e), data={})

    return TxCallResult(
        status=TxStatus.SUCCESS, error='', message='success', data={'gas': estimated_gas}
    )


def estimate_gas(web3: Web3, method: ContractFunction, opts: TxParams) -> int:
    try:
        block_gas_limit = get_block_gas_limit(web3)
    except AttributeError:
        block_gas_limit = get_block_gas_limit(web3)

    estimated_gas = method.estimate_gas(opts, block_identifier='latest')
    normalized_estimated_gas = int(estimated_gas * config.DEFAULT_GAS_MULTIPLIER)
    if normalized_estimated_gas > block_gas_limit:
        logger.warning(
            f'Estimate gas for {method.fn_name} - {normalized_estimated_gas} exceeds \
block gas limit, going to use block_gas_limit ({block_gas_limit}) for this transaction'
        )
        return block_gas_limit
    return normalized_estimated_gas


def build_tx_dict(method: ContractFunction, *args: Any, **kwargs: Any) -> TxParams:
    base_fields = compose_base_fields(*args, **kwargs)
    return method.build_transaction(base_fields)


def compose_base_fields(
    nonce: Nonce,
    gas_limit: int,
    gas_price: Wei | None = None,
    max_fee_per_gas: Wei | None = None,
    max_priority_fee_per_gas: Wei | None = None,
    value: Wei = Wei(0),
) -> TxParams:
    fee_fields = TxParams({'gas': gas_limit, 'nonce': nonce, 'value': value})
    if max_priority_fee_per_gas is not None and max_fee_per_gas is not None:
        fee_fields.update(
            {'maxPriorityFeePerGas': max_priority_fee_per_gas, 'maxFeePerGas': max_fee_per_gas}
        )
        fee_fields.update({'type': 2})
    elif gas_price is not None:
        fee_fields.update({'gasPrice': gas_price})
        fee_fields.update({'type': 1})
    return fee_fields


def transaction_from_method(
    method: ContractFunction,
    *,
    multiplier: Optional[float] = None,
    priority: Optional[int] = None,
    **kwargs: Any,
) -> TxParams:
    tx = build_tx_dict(method, **kwargs)
    logger.info(f'Tx: {method.fn_name}, Fields: {tx}, ')
    return tx


def compose_eth_transfer_tx(
    web3: Web3,
    from_address: ChecksumAddress,
    to_address: ChecksumAddress,
    value: Wei,
    **kwargs: Any,
) -> TxParams:
    nonce = get_eth_nonce(web3, from_address)
    base_fields = compose_base_fields(
        nonce=nonce, gas_limit=DEFAULT_ETH_SEND_GAS_LIMIT, value=value, **kwargs
    )
    tx = TxParams({'from': from_address, 'to': to_address, **base_fields})
    return tx


def retry_tx(
    tx: Callable[..., TxRes] | None = None, *, max_retries: int = 3, timeout: int = -1
) -> Callable[..., TxRes | None] | partial[Any]:
    if tx is None:
        return partial(retry_tx, max_retries=3, timeout=timeout)

    @wraps(tx)
    def wrapper(*args: Any, **kwargs: Any) -> TxRes | None:
        return run_tx_with_retry(
            tx, *args, max_retries=max_retries, retry_timeout=timeout, **kwargs
        )

    return wrapper


def run_tx_with_retry(
    transaction: Callable[..., TxRes],
    *args: Any,
    max_retries: int = 3,
    retry_timeout: int = -1,
    raise_for_status: bool = True,
    **kwargs: Any,
) -> TxRes | None:
    attempt = 0
    tx_res = None
    exp_timeout = 1
    error = None
    while attempt < max_retries:
        try:
            tx_res = transaction(*args, raise_for_status=raise_for_status, **kwargs)
            tx_res.raise_for_status()
        except TransactionError as e:
            error = e
            logger.exception('Tx attempt %d/%d failed', attempt + 1, max_retries)

            timeout = exp_timeout if retry_timeout < 0 else exp_timeout
            time.sleep(timeout)
            exp_timeout *= 2
        else:
            error = None
            break
        attempt += 1
    if error is None:
        logger.info(
            'Tx %s completed after %d/%d retries', transaction.__name__, attempt + 1, max_retries
        )
    else:
        logger.error('Tx %s failed after %d retries', transaction.__name__, max_retries)
        if raise_for_status:
            raise error
    if tx_res is not None:
        tx_res.attempts = attempt
    return tx_res



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.ima.mainnet.base_deposit_box import BaseDepositBox
from skale.types.schain import SchainName
from skale.utils.helper import schain_name_to_hash


class DepositBoxERC20(BaseDepositBox):
    @transaction_method
    def add_erc20_token(
        self, schain_name: SchainName, address: ChecksumAddress
    ) -> ContractFunction:
        return self.contract.functions.addERC20TokenByOwner(schain_name, address)

    @transaction_method
    def deposit_erc20(
        self, schain_name: SchainName, address: ChecksumAddress, amount: int
    ) -> ContractFunction:
        return self.contract.functions.depositERC20(schain_name, address, amount)

    @transaction_method
    def deposit_erc20_direct(
        self, schain_name: SchainName, address: ChecksumAddress, amount: int, receiver: int
    ) -> ContractFunction:
        return self.contract.functions.depositERC20Direct(schain_name, address, amount, receiver)

    @transaction_method
    def set_big_transfer_value(
        self, schain_name: SchainName, token: int, value: int
    ) -> ContractFunction:
        return self.contract.functions.setBigTransferValue(schain_name, token, value)

    @transaction_method
    def set_big_transfer_delay(self, schain_name: SchainName, delay: int) -> ContractFunction:
        return self.contract.functions.setBigTransferDelay(schain_name, delay)

    @transaction_method
    def set_arbitrage_duration(self, schain_name: SchainName, delay: int) -> ContractFunction:
        return self.contract.functions.setArbitrageDuration(schain_name, delay)

    @transaction_method
    def trust_receiver(self, schain_name: SchainName, address: ChecksumAddress) -> ContractFunction:
        return self.contract.functions.trustReceiver(schain_name, address)

    def is_receiver_trusted(self, schain_name: SchainName, address: ChecksumAddress) -> bool:
        return self.contract.functions.isReceiverTrusted(
            schain_name_to_hash(schain_name), address
        ).call()

    def arbiter_role(self) -> bytes:
        return self.contract.functions.ARBITER_ROLE().call()

    def is_token_added(self, schain_name: SchainName, erc20_on_mainnet: ChecksumAddress) -> bool:
        return self.contract.functions.getSchainToERC20(schain_name, erc20_on_mainnet).call()



"""SKALE node rotation test"""

import logging

import pytest

from skale.schain_config.rotation_history import get_previous_schain_groups
from skale.utils.contracts_provision import DEFAULT_SCHAIN_NAME
from skale.utils.contracts_provision.main import (
    add_test4_schain_type,
    cleanup_nodes_schains,
    create_schain,
)
from tests.rotation_history.utils import fail_dkg, remove_node, rotate_node, run_dkg, set_up_nodes

logger = logging.getLogger(__name__)


def test_get_previous_node_no_node(skale):
    assert skale.node_rotation.get_previous_node(DEFAULT_SCHAIN_NAME, 0) is None


@pytest.fixture
def four_node_schain(skale, validator):
    nodes, skale_instances = set_up_nodes(skale, 4, no_zero_id=False)
    add_test4_schain_type(skale)
    try:
        name = create_schain(
            skale,
            schain_type=2,  # test4 type
            random_name=True,
        )
        yield nodes, skale_instances, name
    finally:
        cleanup_nodes_schains(skale)


def test_rotation_history(skale, four_node_schain):
    nodes, skale_instances, name = four_node_schain
    group_index = skale.web3.keccak(text=name)

    run_dkg(nodes, skale_instances, group_index, rotation_id=0)

    group_ids_0 = skale.schains_internal.node_ids_for_schain(name)

    exiting_node_index = 1
    exiting_node_id = nodes[exiting_node_index]['node_id']
    rotate_node(skale, group_index, nodes, skale_instances, exiting_node_index, rotation_id=1)

    previous_node_id = skale.node_rotation.get_previous_node(
        name, nodes[exiting_node_index]['node_id']
    )
    assert previous_node_id == exiting_node_id

    group_ids_1 = skale.schains_internal.node_ids_for_schain(name)

    exiting_node_index = 1
    exiting_node_id = nodes[exiting_node_index]['node_id']
    rotate_node(skale, group_index, nodes, skale_instances, exiting_node_index, rotation_id=2)

    previous_node_id = skale.node_rotation.get_previous_node(
        name, nodes[exiting_node_index]['node_id']
    )
    assert previous_node_id == exiting_node_id

    group_ids_2 = skale.schains_internal.node_ids_for_schain(name)

    exiting_node_index = 2
    exiting_node_id = nodes[exiting_node_index]['node_id']
    rotate_node(skale, group_index, nodes, skale_instances, exiting_node_index, rotation_id=3)

    previous_node_id = skale.node_rotation.get_previous_node(
        name, nodes[exiting_node_index]['node_id']
    )
    assert previous_node_id == exiting_node_id

    group_ids_3 = skale.schains_internal.node_ids_for_schain(name)

    exiting_node_index = 3
    exiting_node_id = nodes[exiting_node_index]['node_id']
    rotate_node(skale, group_index, nodes, skale_instances, exiting_node_index, rotation_id=4)

    previous_node_id = skale.node_rotation.get_previous_node(
        name, nodes[exiting_node_index]['node_id']
    )
    assert previous_node_id == exiting_node_id

    group_ids_4 = skale.schains_internal.node_ids_for_schain(name)

    exiting_node_index = 1
    exiting_node_id = nodes[exiting_node_index]['node_id']
    rotate_node(skale, group_index, nodes, skale_instances, exiting_node_index, rotation_id=5)

    previous_node_id = skale.node_rotation.get_previous_node(
        name, nodes[exiting_node_index]['node_id']
    )
    assert previous_node_id == exiting_node_id

    group_ids_5 = skale.schains_internal.node_ids_for_schain(name)

    node_groups = get_previous_schain_groups(skale, name)

    assert len(node_groups) == 6
    assert set(node_groups[0]['nodes'].keys()) == set(group_ids_0)
    assert set(node_groups[1]['nodes'].keys()) == set(group_ids_1)
    assert set(node_groups[2]['nodes'].keys()) == set(group_ids_2)
    assert set(node_groups[3]['nodes'].keys()) == set(group_ids_3)
    assert set(node_groups[4]['nodes'].keys()) == set(group_ids_4)
    assert set(node_groups[5]['nodes'].keys()) == set(group_ids_5)


def test_rotation_history_no_rotations(skale, four_node_schain):
    _, _, name = four_node_schain
    node_groups = get_previous_schain_groups(skale, name)
    group_ids = skale.schains_internal.node_ids_for_schain(name)

    assert len(node_groups) == 1
    assert set(node_groups[0]['nodes'].keys()) == set(group_ids)


def test_rotation_history_single_rotation(skale, four_node_schain):
    nodes, skale_instances, name = four_node_schain
    group_index = skale.web3.keccak(text=name)

    run_dkg(nodes, skale_instances, group_index, rotation_id=0)

    group_ids_0 = skale.schains_internal.node_ids_for_schain(name)

    exiting_node_index = 1
    rotate_node(skale, group_index, nodes, skale_instances, exiting_node_index, rotation_id=1)

    group_ids_1 = skale.schains_internal.node_ids_for_schain(name)

    node_groups = get_previous_schain_groups(skale, name)

    assert len(node_groups) == 2
    assert set(node_groups[0]['nodes'].keys()) == set(group_ids_0)
    assert set(node_groups[1]['nodes'].keys()) == set(group_ids_1)


@pytest.mark.parametrize(
    'first_node_index_to_exit,failed_node_index,second_node_index_to_exit',
    [(1, 2, 1), (1, 2, 3), (2, 2, 2)],
)
def test_rotation_history_failed_dkg(
    skale, four_node_schain, first_node_index_to_exit, failed_node_index, second_node_index_to_exit
):
    nodes, skale_instances, name = four_node_schain
    group_index = skale.web3.keccak(text=name)
    assert not skale.dkg.is_node_broadcasted(group_index, nodes[0]['node_id'])

    run_dkg(nodes, skale_instances, group_index, rotation_id=0)
    assert skale.dkg.is_node_broadcasted(group_index, nodes[0]['node_id'])

    group_ids_0 = skale.schains_internal.node_ids_for_schain(name)

    rotate_node(
        skale,
        group_index,
        nodes,
        skale_instances,
        first_node_index_to_exit,
        do_dkg=False,
        rotation_id=1,
    )

    group_ids_1 = skale.schains_internal.node_ids_for_schain(name)

    failed_node_id = nodes[failed_node_index]['node_id']
    fail_dkg(skale, nodes, skale_instances, group_index, failed_node_index, rotation_id=1)

    group_ids_2 = skale.schains_internal.node_ids_for_schain(name)

    exiting_node_index = second_node_index_to_exit
    exiting_node_id = nodes[exiting_node_index]['node_id']
    rotate_node(skale, group_index, nodes, skale_instances, exiting_node_index, rotation_id=3)

    previous_node_id = skale.node_rotation.get_previous_node(
        name, nodes[exiting_node_index]['node_id']
    )
    assert previous_node_id == exiting_node_id

    group_ids_3 = skale.schains_internal.node_ids_for_schain(name)

    remove_node(skale, failed_node_id)

    node_groups = get_previous_schain_groups(skale, name)

    assert len(node_groups) == 4
    assert set(node_groups[0]['nodes'].keys()) == set(group_ids_0)
    assert set(node_groups[1]['nodes'].keys()) == set(group_ids_1)
    assert set(node_groups[2]['nodes'].keys()) == set(group_ids_2)
    assert set(node_groups[3]['nodes'].keys()) == set(group_ids_3)

    assert node_groups[0]['finish_ts']
    assert node_groups[0]['bls_public_key']

    # no keys and finish_ts because it's the group that failed DKG
    assert not node_groups[1]['finish_ts']
    assert not node_groups[1]['bls_public_key']

    assert node_groups[2]['finish_ts']
    assert node_groups[2]['bls_public_key']

    # no finish_ts because it's the current group
    assert not node_groups[3]['finish_ts']
    assert node_groups[3]['bls_public_key']



from unittest import mock

import pytest
import web3

from skale.transactions.exceptions import ChainIdError
from skale.utils.web3_utils import init_web3
from skale.wallets import Web3Wallet
from skale.wallets.common import ensure_chain_id
from tests.constants import ENDPOINT, ETH_PRIVATE_KEY


def get_tx_dict():
    return {
        'to': '0x0000000000000000000000000000000000000000',
        'value': 0,
        'gasPrice': 1,
        'gas': 22000,
        'nonce': 0,
    }


def test_ensure_chain_id(skale):
    tx_dict = get_tx_dict()
    ensure_chain_id(tx_dict, skale.web3)
    assert isinstance(tx_dict['chainId'], int)
    assert tx_dict['chainId'] == skale.web3.eth.chain_id


def test_ensure_chain_id_fail(skale):
    tx_dict = get_tx_dict()
    with mock.patch.object(web3.eth.Eth, 'chain_id', new=None):
        with pytest.raises(ChainIdError):
            ensure_chain_id(tx_dict, skale.web3)


def test_ensure_chain_id_web3_wallet():
    tx_dict = get_tx_dict()
    web3 = init_web3(ENDPOINT)
    wallet = Web3Wallet(ETH_PRIVATE_KEY, web3)
    wallet.sign(tx_dict)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
from typing import Tuple, cast

from eth_account.datastructures import SignedMessage, SignedTransaction
from eth_typing import ChecksumAddress, HexStr
from sgx import SgxClient
from web3 import Web3
from web3.exceptions import Web3Exception
from web3.types import TxParams, TxReceipt, _Hash32

import skale.config as config
from skale.transactions.exceptions import TransactionNotSentError, TransactionNotSignedError
from skale.utils.web3_utils import (
    DEFAULT_BLOCKS_TO_WAIT,
    MAX_WAITING_TIME,
    get_eth_nonce,
    wait_for_receipt,
)
from skale.wallets.common import BaseWallet, MessageNotSignedError, ensure_chain_id

logger = logging.getLogger(__name__)


class SgxWallet(BaseWallet):
    def __init__(
        self,
        sgx_endpoint: str,
        web3: Web3,
        key_name: str | None = None,
        path_to_cert: str | None = None,
    ):
        self.sgx_client = SgxClient(sgx_endpoint, path_to_cert=path_to_cert)
        self._web3 = web3
        if key_name is None:
            self._key_name, self._address, self._public_key = self._generate()
        else:
            self._key_name = key_name
            self._address, self._public_key = self._get_account(key_name)

    def sign(self, tx_dict: TxParams) -> SignedTransaction:
        if tx_dict.get('nonce') is None:
            tx_dict['nonce'] = get_eth_nonce(self._web3, self._address)
        ensure_chain_id(tx_dict, self._web3)
        try:
            return cast(SignedTransaction, self.sgx_client.sign(tx_dict, self.key_name))
        except Exception as e:
            raise TransactionNotSignedError(e)

    def sign_and_send(
        self,
        tx_dict: TxParams,
        multiplier: float | None = config.DEFAULT_GAS_MULTIPLIER,
        priority: int | None = config.DEFAULT_PRIORITY,
        method: str | None = None,
    ) -> HexStr:
        signed_tx = self.sign(tx_dict)
        try:
            return Web3.to_hex(self._web3.eth.send_raw_transaction(signed_tx.raw_transaction))
        except (ValueError, Web3Exception) as e:
            raise TransactionNotSentError(e)

    def sign_hash(self, unsigned_hash: str) -> SignedMessage:
        if unsigned_hash.startswith('0x'):
            unsigned_hash = unsigned_hash[2:]

        body = bytes.fromhex(unsigned_hash)
        header = b'\x19Ethereum Signed Message:\n32'
        normalized_hash = header + body
        hash_to_sign = Web3.keccak(hexstr='0x' + normalized_hash.hex())
        chain_id = None
        try:
            return cast(
                SignedMessage, self.sgx_client.sign_hash(hash_to_sign, self._key_name, chain_id)
            )
        except Exception as e:
            raise MessageNotSignedError(e)

    @property
    def address(self) -> ChecksumAddress:
        return self._address

    @property
    def public_key(self) -> str:
        return self._public_key

    @property
    def key_name(self) -> str:
        return self._key_name

    def _generate(self) -> Tuple[str, ChecksumAddress, str]:
        key = self.sgx_client.generate_key()
        return key.name, Web3.to_checksum_address(key.address), key.public_key

    def _get_account(self, key_name: str) -> Tuple[ChecksumAddress, str]:
        account = self.sgx_client.get_account(key_name)
        return Web3.to_checksum_address(account.address), account.public_key

    def wait(
        self,
        tx_hash: _Hash32,
        blocks_to_wait: int = DEFAULT_BLOCKS_TO_WAIT,
        timeout: int = MAX_WAITING_TIME,
    ) -> TxReceipt:
        return wait_for_receipt(self._web3, tx_hash, timeout=timeout)



import random

import pytest

from skale.transactions.result import DryRunRevertError
from tests.constants import NEW_DELTA_PERIOD, NEW_REWARD_PERIOD


def test_get_set_periods(skale):
    skale.constants_holder.set_check_time(10, wait_for=True)
    tx_res = skale.constants_holder.set_periods(NEW_REWARD_PERIOD, NEW_DELTA_PERIOD, wait_for=True)
    assert tx_res.receipt['status'] == 1
    reward_period = skale.constants_holder.get_reward_period()
    delta_period = skale.constants_holder.get_delta_period()
    assert reward_period == NEW_REWARD_PERIOD
    assert delta_period == NEW_DELTA_PERIOD


def test_get_set_check_time(skale):
    new_check_time = 100
    tx_res = skale.constants_holder.set_check_time(new_check_time, wait_for=True)
    assert tx_res.receipt['status'] == 1
    res = skale.constants_holder.get_check_time()
    assert res == new_check_time


def test_get_set_latency(skale):
    new_latency = 1000
    tx_res = skale.constants_holder.set_latency(new_latency, wait_for=True)
    assert tx_res.receipt['status'] == 1
    res = skale.constants_holder.get_latency()
    assert res == new_latency


def test_get_set_launch_timestamp(skale):
    launch_ts = skale.constants_holder.get_launch_timestamp()
    assert isinstance(launch_ts, int)
    with pytest.raises(DryRunRevertError):
        skale.constants_holder.set_launch_timestamp(launch_ts, wait_for=True)


def test_get_set_rotation_delay(skale):
    rotation_delay = skale.constants_holder.get_rotation_delay()
    assert isinstance(rotation_delay, int) and rotation_delay > 0
    new_rotation_delay = 1000
    skale.constants_holder.set_rotation_delay(new_rotation_delay, wait_for=True)
    rotation_delay = skale.constants_holder.get_rotation_delay()
    assert rotation_delay == new_rotation_delay


def test_get_first_delegation_month(skale):
    fdm = skale.constants_holder.get_first_delegation_month()
    assert fdm == 0


def test_get_set_complaint_timelimit(skale):
    new_dkg_timeout = random.randint(100, 100000)
    dkg_timeout_before = skale.constants_holder.get_dkg_timeout()
    skale.constants_holder.set_complaint_timelimit(new_dkg_timeout, wait_for=True)
    dkg_timeout_after = skale.constants_holder.get_dkg_timeout()
    assert dkg_timeout_after != dkg_timeout_before
    assert dkg_timeout_after == new_dkg_timeout



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
"""SchainsInternal.sol functions"""

import functools
from typing import TYPE_CHECKING

from eth_typing import ChecksumAddress

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.types.node import NodeId
from skale.types.schain import Schain, SchainHash, SchainName
from skale.utils.helper import schain_name_to_hash

if TYPE_CHECKING:
    from web3.contract.contract import ContractFunction

    from skale.contracts.manager.schains import SChains


class SChainsInternal(SkaleManagerContract):
    """Wrapper for some of the SchainsInternal.sol functions"""

    @property
    @functools.lru_cache()
    def schains(self) -> 'SChains':
        return self.skale.schains

    def get_raw(self, name: SchainHash) -> Schain:
        return Schain(*self.contract.functions.schains(name).call())

    def all_schain_hashes(self) -> list[SchainHash]:
        return [
            SchainHash(schain_hash) for schain_hash in self.contract.functions.getSchains().call()
        ]

    def number_of_schains(self) -> int:
        return int(self.contract.functions.numberOfSchains().call())

    def schain_list_size(self, account: ChecksumAddress) -> int:
        return int(self.contract.functions.getSchainListSize(account).call({'from': account}))

    def schain_hash_by_index_for_owner(self, account: ChecksumAddress, index: int) -> SchainHash:
        return SchainHash(self.contract.functions.schainIndexes(account, index).call())

    def node_ids_for_schain(self, name: SchainName) -> list[NodeId]:
        schain_hash = schain_name_to_hash(name)
        return self.node_ids_for_schain_hash(schain_hash)

    def node_ids_for_schain_hash(self, schain_hash: SchainHash) -> list[NodeId]:
        return [
            NodeId(node) for node in self.contract.functions.getNodesInGroup(schain_hash).call()
        ]

    def connected_node_ids(self, node_id: NodeId) -> list[NodeId]:
        """Gets all schain hashes for node, retrieves node ids for each schain and returns a list
        with unique node ids for all schains on the node.

        :returns: List of unique node ids
        :rtype: list[NodeId]
        """
        schains = self.schain_hashes_for_node(node_id)
        unique_node_ids: set[NodeId] = set()
        for schain_hash in schains:
            if self.is_empty_schain_hash(schain_hash):
                continue
            group_node_ids = self.node_ids_for_schain_hash(schain_hash)
            unique_node_ids.update(group_node_ids)
        return list(unique_node_ids)

    def schain_hashes_for_node(self, node_id: NodeId) -> list[SchainHash]:
        return [
            SchainHash(schain)
            for schain in self.contract.functions.getSchainHashesForNode(node_id).call()
        ]

    def is_schain_exist(self, name: SchainName) -> bool:
        schain_hash = schain_name_to_hash(name)
        return bool(self.contract.functions.isSchainExist(schain_hash).call())

    def is_empty_schain_hash(self, schain_hash: SchainHash) -> bool:
        return all(b == 0 for b in schain_hash)

    def active_schain_hashes_for_node(self, node_id: NodeId) -> list[SchainHash]:
        return [
            SchainHash(schain)
            for schain in self.contract.functions.getActiveSchains(node_id).call()
        ]

    def number_of_schain_types(self) -> int:
        return int(self.contract.functions.numberOfSchainTypes().call())

    @transaction_method
    def add_schain_type(self, part_of_node: int, number_of_nodes: int) -> 'ContractFunction':
        return self.contract.functions.addSchainType(part_of_node, number_of_nodes)

    def current_generation(self) -> int:
        return int(self.contract.functions.currentGeneration().call())

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> 'ContractFunction':
        return self.contract.functions.grantRole(role, address)

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.hasRole(role, address).call())

    def schain_type_manager_role(self) -> bytes:
        return bytes(self.contract.functions.SCHAIN_TYPE_MANAGER_ROLE().call())

    def debugger_role(self) -> bytes:
        return bytes(self.contract.functions.DEBUGGER_ROLE().call())

    def generation_manager_role(self) -> bytes:
        return bytes(self.contract.functions.GENERATION_MANAGER_ROLE().call())

    @transaction_method
    def new_generation(self) -> 'ContractFunction':
        return self.contract.functions.newGeneration()

    def check_exception(self, schain_name: SchainName, node_id: NodeId) -> bool:
        schain_hash = schain_name_to_hash(schain_name)
        return bool(self.contract.functions.checkException(schain_hash, node_id).call())
